#
from scipy import optimize
import matplotlib.pyplot as plt
import tqdm
import pickle
from scipy import stats
from scipy import sparse
import pandas as pd
import numpy as np
import time

from robustlink import utils

def set_venn_scale(ax, true_area, ref_area):
    s = np.sqrt(ref_area/true_area)
    ax.set_xlim(-s, s)
    ax.set_ylim(-s, s)
    return

def turn_cluster_labels_to_knn(cluster_labels, uniq_labels):
    """
    """

    clst_idx = utils.get_index_from_array(uniq_labels, cluster_labels)
    m, n = len(clst_idx), len(uniq_labels)
    _i = np.arange(m)
    _j = clst_idx
    _data = [1]*m
    knn = sparse.coo_matrix((_data, (_i, _j)), shape=(m, n))
    return knn

def row_dot_product_norm_by_numcol(X_zscore, Y_zscore, x_idx, y_idx,
                                   chunksize=100000, verbose_level=100000):
    """compute (X_zscore[x_idx]*Y_zscore[y_idx]).mean(axis=1)
    correlation values given matched x_idx and y_idx...
    """
    ti = time.time()

    assert len(x_idx) == len(y_idx)
    num_pairs = len(x_idx)
    corrs = []
    for pair_idx in utils.chunks(np.arange(num_pairs), chunksize):
        if verbose_level and pair_idx[0] % verbose_level == 0:
            print(pair_idx[0], time.time()-ti)

        _res = (X_zscore[x_idx[pair_idx]] *
                Y_zscore[y_idx[pair_idx]]).mean(axis=1)
        corrs.append(_res)
    corrs = np.hstack(corrs)
    return corrs

def compute_enh_gene_corrs(gc_rna, ec_mccg,
                           genes, enhancers,
                           pairs_gene, pairs_enh,
                           enhancer_groups=[],
                           output_file='', corr_type='pearsonr', shuff_enhs=False, **kwargs,
                           ):
    """Compute correlations and 2 shuffled correlations
    inputs:
        gc_rna: enh-by-cell RNA matrix
        ec_mccg: enh-by-cell mC matrix

        genes: gene list in the order of the mat gc_rna
        enhancers: enhancer list in the order of the mat ec_mccg
        pairs_gene: pairs to evalute (gene)
        pairs_enh: pairs to evaluate (enh)
        (referencing genes and enhancers,
        if a gene/enh is not in the genes/enhancers list, do not compute the correlations
        )
        enhancer_groups: an array with the same length as enhancers (and in the same order as enhancers)
    outputs:
        to_correlate # pairs computed correlations
        corrs, corrs_shuffled, corrs_shuffled_cells
    """
    print("{} chosen!".format(corr_type))
    if corr_type == 'pearsonr':
        pass
    elif corr_type == 'spearmanr':
        # ranking - across cells
        gc_rna = pd.DataFrame(np.array(gc_rna)).rank(axis=1).values
        ec_mccg = pd.DataFrame(np.array(ec_mccg)).rank(axis=1).values

    gc_rna_zscore = stats.zscore(
        np.array(gc_rna), axis=1, ddof=0, nan_policy='propagate')
    ec_mccg_zscore = stats.zscore(
        np.array(ec_mccg), axis=1, ddof=0, nan_policy='propagate')

    # correlate e-g according to a e-g table
    gene_idx = utils.get_index_from_array(genes, pairs_gene)
    enh_idx = utils.get_index_from_array(
        enhancers, pairs_enh)  # be careful here!
    to_correlate = ~np.logical_or(gene_idx == -1, enh_idx == -1)

    gene_idx = gene_idx[to_correlate]
    enh_idx = enh_idx[to_correlate]

    # corr
    corrs = row_dot_product_norm_by_numcol(
        gc_rna_zscore, ec_mccg_zscore, gene_idx, enh_idx, **kwargs)

    # corr shuffled cells
    corrs_shuffled_cells = row_dot_product_norm_by_numcol(
        gc_rna_zscore[:, np.random.permutation(gc_rna_zscore.shape[1])],
        ec_mccg_zscore,
        gene_idx, enh_idx, **kwargs)

    # corr shuffled genes (break up the pairs)
    gene_idx_uniq = np.unique(gene_idx)
    shuff_genes = {
        gene: gene_shuff for gene, gene_shuff in
        zip(gene_idx_uniq,
            gene_idx_uniq[np.random.permutation(len(gene_idx_uniq))])
    }
    gene_idx_shuff = np.array([shuff_genes[gene] for gene in gene_idx])

    corrs_shuffled = row_dot_product_norm_by_numcol(
        gc_rna_zscore,
        ec_mccg_zscore,
        gene_idx_shuff, enh_idx, **kwargs)

    if not shuff_enhs:
        output = (to_correlate, corrs, corrs_shuffled, corrs_shuffled_cells)
        # save and return
        if output_file:
            with open(output_file, 'wb') as fh:
                pickle.dump(output, fh)
        return output
    else:
        # corr shuffled enhancers
        # shuffle enhancers within each group
        if len(enhancer_groups) > 0:
            # control for the groupings
            enh_idx_uniq = np.unique(enh_idx) # unique enhancers
            enh_idx_groups = enhancer_groups[enh_idx] # enh_idx - enh_idx_groups
            enh_grps_uniq = np.unique(enh_idx_groups) # unique groups
            shuff_enhs = {}
            for enh_grp in enh_grps_uniq:
                enh_idx_grp = enh_idx[enh_idx_groups==enh_grp]
                enh_idx_grp_uniq = np.unique(enh_idx_grp)
                shuff_enhs.update({
                    enh: enh_shuff for enh, enh_shuff in
                    zip(enh_idx_grp_uniq,
                        enh_idx_grp_uniq[np.random.permutation(len(enh_idx_grp_uniq))])
                })
            enh_idx_shuff = np.array([shuff_enhs[enh] for enh in enh_idx])
            corrs_shuffled_enhs_bygroups = row_dot_product_norm_by_numcol(
                gc_rna_zscore,
                ec_mccg_zscore,
                gene_idx, enh_idx_shuff, **kwargs)
        else:
            corrs_shuffled_enhs_bygroups = 0

        # shuffle all enhancers
        enh_idx_uniq = np.unique(enh_idx)
        shuff_enhs = {
            enh: enh_shuff for enh, enh_shuff in
            zip(enh_idx_uniq,
                enh_idx_uniq[np.random.permutation(len(enh_idx_uniq))])
        }
        enh_idx_shuff = np.array([shuff_enhs[enh] for enh in enh_idx])
        corrs_shuffled_enhs = row_dot_product_norm_by_numcol(
            gc_rna_zscore,
            ec_mccg_zscore,
            gene_idx, enh_idx_shuff, **kwargs)

        output = (to_correlate, corrs, corrs_shuffled, corrs_shuffled_cells, 
                corrs_shuffled_enhs,
                corrs_shuffled_enhs_bygroups,
                )
        # save and return
        if output_file:
            with open(output_file, 'wb') as fh:
                pickle.dump(output, fh)
        return output

DEBUG_MODE = False
def get_r_threshold_smart(bins, fdr, fdr_threshold, side='left'):
    """given fdr function (bins, fdr) and fdr_threshold
    get r threshold for a given fdr_treshold
    """
    # get r_threshold
    # remove nan
    flag = 1  # default no solution

    isnan = np.isnan(fdr)
    _y = fdr[~isnan]
    _x = bins[1:][~isnan]

    # find r threshold
    def f(_x_func): return np.interp(_x_func, _x, _y) - fdr_threshold

    # solve
    if side == 'left':
        bins_sub = bins[bins < 0]
    elif side == 'right':
        bins_sub = bins[bins > 0]
    else:
        raise ValueError("left or right")

    r_min = bins_sub[np.argmin(f(bins_sub))]
    if f(r_min)*f(0) <0:
        sol = optimize.root_scalar(f, bracket=(r_min, 0))
        if sol:
            r_threshold = sol.root
            flag = 0

    if flag == 1:
        r_threshold = np.nan
        print("failed to detect r_threshold:")

    # if DEBUG_MODE and flag == 1:
    if DEBUG_MODE:
        fig, ax = plt.subplots()
        ax.plot(bins, f(bins))
        if flag == 0:
            print(r_threshold, f(r_threshold))
            ax.scatter([r_threshold], [f(r_threshold)])
        # ax2 = ax.twinx()
        # ax2.plot(bins[1:], f(bins[1:]) - f(bins[:-1]), color='C1')
        plt.show()

    return r_threshold


def cumfrac_to_pval(cumfracs, pval_type):
    """Null hypothese
    """
    if pval_type == 'left':
        pvals = cumfracs
    elif pval_type == 'right':
        pvals = 1 - cumfracs
    elif pval_type == 'both':
        alpha = np.minimum(cumfracs, 1-cumfracs)
        pvals = 2*alpha
    else:
        raise ValueError
    return pvals


def cumfrac_to_pval_obs_simple(cumfracs, pval_type):
    """Observations
    Assuming null is symmetric given the bin range
    """
    if pval_type == 'left':
        pvals = cumfracs
    elif pval_type == 'right':
        pvals = 1 - cumfracs
    elif pval_type == 'both':
        alpha = np.minimum(cumfracs, 1-cumfracs)
        alpha_star = alpha[::-1]  # (this implies bins_conj == bins[::-1])
        pvals = alpha + alpha_star
    else:
        raise ValueError
    return pvals


def cumfrac_to_pval_obs_general(bins, cumfracs, cumfracs_null, pval_type):
    """Observations
    """
    if pval_type == 'left':
        pvals = cumfracs
    elif pval_type == 'right':
        pvals = 1 - cumfracs
    elif pval_type == 'both':
        alpha_null = np.minimum(cumfracs_null, 1-cumfracs_null)
        alpha = np.minimum(cumfracs, 1-cumfracs)
        # look for bins_conj such that alpha_null(bins_conj) == alpha_null(bins)
        # bins_conj = get_conjugate(bins, alpha_null)
        # alpha_star = # alpha(bins_conj)
        # pvals = alpha + alpha_star
        raise ValueError("Not yet implemented")
    else:
        raise ValueError
    return pvals


def get_significance_stats(
            pairs,
            corrs, corrs_shuffled, corrs_shuffled_cells,
            pval_type_shuffled, pval_type_shuffled_cells,
            bins=np.linspace(-1, 1,101),
            distance_threshold=1e5, fdr_threshold=0.2,
            positive_side=False,
            return_pval=False,
            return_cdf=False,
        ):
    """assuming significant negative correlation
    pairs should be a df with at least 3 columns: gene, enh, dist
    """
    # align all to negative side
    assert pval_type_shuffled in ['left', 'both', 'right']  # one/two side test
    assert pval_type_shuffled_cells == 'both'  # two sides test

    if positive_side:
        # prevent modifying the original one
        corrs = corrs.copy()
        corrs_shuffled = corrs_shuffled.copy()
        corrs_shuffled_cells = corrs_shuffled_cells.copy()

        corrs *= (-1)
        corrs_shuffled *= (-1)
        corrs_shuffled_cells *= (-1)

    # dists
    dists = pairs['dist'].values
    label_cond = dists < distance_threshold
    track = corrs[label_cond]

    # total numbers with the condition
    num_total_pairs = len(pairs[label_cond])
    num_total_genes = len(pairs[label_cond]['gene'].unique())
    num_total_enhs = len(pairs[label_cond]['enh'].unique())

    # hist_shuff
    hist_shuff, _ = np.histogram(corrs_shuffled, bins=bins, density=True)
    cdf_shuff = np.cumsum(hist_shuff)/np.sum(hist_shuff)
    pval_shuff = cumfrac_to_pval(cdf_shuff, pval_type_shuffled)
    # hist_shuff_cells
    hist_shuff_cells, _ = np.histogram(corrs_shuffled_cells, bins=bins, density=True)
    cdf_shuff_cells = np.cumsum(hist_shuff_cells)/np.sum(hist_shuff_cells)
    pval_shuff_cells = cumfrac_to_pval(cdf_shuff_cells, pval_type_shuffled_cells)
    # hist
    hist, _ = np.histogram(track, bins=bins, density=True)
    cdf = np.cumsum(hist)/np.sum(hist)
    # modified on 12/23/2020
    pval_obs_shuff = cumfrac_to_pval_obs_simple(cdf, pval_type_shuffled)
    pval_obs_shuff_cells = cumfrac_to_pval_obs_simple(cdf, pval_type_shuffled_cells)

    # fdr
    fdr_linked = (pval_shuff+1e-7)/(pval_obs_shuff+1e-7)
    fdr_correlated = (pval_shuff_cells+1e-7)/(pval_obs_shuff_cells+1e-7)

    # get all fdrs
    track_fdr_linked = np.interp(track, bins[1:], fdr_linked)
    track_fdr_correlated = np.interp(track, bins[1:], fdr_correlated)

    # get r_threshold
    if pval_type_shuffled == 'both':
        r_threshold_linked = np.nan
        r_threshold_linked_left = get_r_threshold_smart(bins, fdr_linked, fdr_threshold, side='left')
        r_threshold_linked_right = get_r_threshold_smart(bins, fdr_linked, fdr_threshold, side='right')
    else:
        r_threshold_linked = get_r_threshold_smart(bins, fdr_linked, fdr_threshold, side=pval_type_shuffled)
        r_threshold_linked_left = np.nan
        r_threshold_linked_right = np.nan

    r_threshold_correlated_left = get_r_threshold_smart(bins, fdr_correlated, fdr_threshold, side='left')
    r_threshold_correlated_right = get_r_threshold_smart(bins, fdr_correlated, fdr_threshold, side='right')

    # stats
    # all sig info
    linked_table = pairs[label_cond][track_fdr_linked < fdr_threshold]
    num_linked_pairs = len(linked_table)
    num_linked_genes = len(linked_table['gene'].unique())
    num_linked_enhs = len(linked_table['enh'].unique())
    # all sig info
    correlated_table = pairs[label_cond][track_fdr_correlated < fdr_threshold]
    num_correlated_pairs = len(correlated_table)
    num_correlated_genes = len(correlated_table['gene'].unique())
    num_correlated_enhs = len(correlated_table['enh'].unique())

    output = {
        'dist_th': distance_threshold,
                'num_total_pairs': num_total_pairs,
                'num_total_genes': num_total_genes,
                'num_total_enhs': num_total_enhs,

                'r_th_linked': r_threshold_linked,
                'r_th_linked_left': r_threshold_linked_left,
                'r_th_linked_right': r_threshold_linked_right,

                'r_th_correlated_left': r_threshold_correlated_left,
                'r_th_correlated_right': r_threshold_correlated_right,

                'num_linked_pairs': num_linked_pairs,
                'num_linked_genes': num_linked_genes,
                'num_linked_enhs': num_linked_enhs,
                'linked_table': linked_table,

                'num_correlated_pairs': num_correlated_pairs,
                'num_correlated_genes': num_correlated_genes,
                'num_correlated_enhs': num_correlated_enhs,
                'correlated_table': correlated_table,
        }

    if return_pval:
        output['bins'] = bins
        output['linked_pval'] = pval_shuff
        output['correlated_pval'] = pval_shuff_cells
    if return_cdf:
        output['bins'] = bins
        output['linked_cdf'] = cdf_shuff
        output['correlated_cdf'] = cdf_shuff_cells
    return output

def get_corr_stats(iterator_both, enhancer_gene_to_eval, col_orders,
    bins=np.linspace(-1, 1,101),
    distance_threshold=1e5,
                   fdr_threshold=0.2,
                   r_min=-1,
                   r_max=0,
    ):
    """Wrap-up the routine of calculating corr stats, and compare between mC, ATAC, and both
    used in jupyter notebook visualizations
    """
    pval_type_shuffled, pval_type_shuffled_cells = 'left', 'both'
    positive_sides = [False, True]
    res = []
    for idx, row in tqdm.tqdm(iterator_both.iterrows()):
        fname_mc, fname_atac = (row['fname_mc'], row['fname_atac'],)

        res_2cases = []
        for i, (label, fname, positive_side) in enumerate(zip(
            ['mc', 'atac'],
                            [fname_mc, fname_atac],
                            [False, True],
            )):
            try:
                with open(fname, 'rb') as fh:
                    to_correlate, corrs, corrs_shuffled, corrs_shuffled_cells = pickle.load(fh)
            except:
                continue

            pairs = enhancer_gene_to_eval[to_correlate].copy()
            label_cond = pairs['dist'].values < distance_threshold

            res_1case = get_significance_stats(
                pairs,
                corrs, corrs_shuffled, corrs_shuffled_cells,
                pval_type_shuffled, pval_type_shuffled_cells,
                bins=bins,
                distance_threshold=distance_threshold, fdr_threshold=fdr_threshold,
                positive_side=positive_side,
                return_cdf=False,
                return_pval=False,
            )

            # record
            res_1case['id_total_pairs'] = pairs[label_cond].index.values

            if isinstance(res_1case['linked_table'], pd.DataFrame):
                res_1case['id_linked_pairs'] = res_1case['linked_table'].index.values
            else:
                res_1case['id_linked_pairs'] = np.array([])

            if isinstance(res_1case['correlated_table'], pd.DataFrame):
                res_1case['id_correlated_pairs'] = res_1case['correlated_table'].index.values
            else:
                res_1case['id_correlated_pairs'] = np.array([])

            res_1case = pd.Series(res_1case)[col_orders]
            res_1case.index = res_1case.index + '_{}'.format(label)
            res_2cases.append(res_1case)

        # coommon
        res_2cases = pd.concat(res_2cases)

        # linked both
        common_pair_ids = np.intersect1d(res_2cases['id_linked_pairs_mc'],
                                         res_2cases['id_linked_pairs_atac'],
                                         )
        common_gene_ids = enhancer_gene_to_eval.loc[common_pair_ids, 'gene'].unique()
        common_enh_ids = enhancer_gene_to_eval.loc[common_pair_ids, 'enh'].unique()
        res_2cases['num_linked_pairs_both'] = len(common_pair_ids)
        res_2cases['num_linked_genes_both'] = len(common_gene_ids)
        res_2cases['num_linked_enhs_both'] = len(common_enh_ids)

        # correlated both
        common_pair_ids = np.intersect1d(res_2cases['id_correlated_pairs_mc'],
                                         res_2cases['id_correlated_pairs_atac'],
                                         )
        common_gene_ids = enhancer_gene_to_eval.loc[common_pair_ids, 'gene'].unique()
        common_enh_ids = enhancer_gene_to_eval.loc[common_pair_ids, 'enh'].unique()
        res_2cases['num_correlated_pairs_both'] = len(common_pair_ids)
        res_2cases['num_correlated_genes_both'] = len(common_gene_ids)
        res_2cases['num_correlated_enhs_both'] = len(common_enh_ids)

        # total both
        common_pair_ids = np.intersect1d(res_2cases['id_total_pairs_mc'],
                                         res_2cases['id_total_pairs_atac'],
                                         )
        common_gene_ids = enhancer_gene_to_eval.loc[common_pair_ids, 'gene'].unique()
        common_enh_ids = enhancer_gene_to_eval.loc[common_pair_ids, 'enh'].unique()
        res_2cases['num_total_pairs_both'] = len(common_pair_ids)
        res_2cases['num_total_genes_both'] = len(common_gene_ids)
        res_2cases['num_total_enhs_both'] = len(common_enh_ids)

        res.append(pd.Series(res_2cases))
    res = pd.DataFrame(res, index=iterator_both.index)
    print(res.shape)
    return res



#!/usr/bin/env python3
"""SingleCellFusion main rontine"""

from .__init__ import *

import collections
import argparse
import anndata

from . import basic_utils
from . import SCF_utils

def add_args(parser):
    """
    """
    # parser.add_argument("-c", "--config_py", help="Configuration file", required=True)

    required = parser.add_argument_group('required')
    optional = parser.add_argument_group('optional') 
    advanced = parser.add_argument_group('advanced')

    required.add_argument("-i", "--data_dir", help='data dir', required=True)
    required.add_argument("-o", "--outdir", help='outdir', required=True)
    required.add_argument(
        "-id", "--input_datasets", 
        type=str,
        nargs="+",
        required=True,
        help='''(list of str)
             list of .h5ad files (AnnData) in the data_dir 
             '''
    )
    required.add_argument(
        "-im", "--input_modalities", 
        type=str,
        nargs="+",
        required=True,
        help='''(list of str)
             Data modalities chosen from 'rna', 'atac', or 'mc'. This should be 
             listed in the same order as input_datasets. 
             '''
    )
    required.add_argument(
        "-fd", "--feature_datasets", 
        type=str,
        nargs="+",
        required=True,
        help='''(list of str)
             Dataset(s) whose features all other datasets will impute into.
             This should be a subset of --input_datasets.
             Enter multiple datasets as a space-separated list of filenames.
             '''
    )

    optional.add_argument("-tag", "--nametag", help='output files will contain this tag', type=str, default='scfusion')
    optional.add_argument("--ka_smooth", type=int, default=30)
    optional.add_argument("--knn", help=">1", type=int, default=30)
    optional.add_argument("-s", "--subsample_fraction", help="0~1", type=float, default=1)
    optional.add_argument("-sn", "--subsample_times", help=">1", type=int, default=1)
    # constraint kNN across modalities
    optional.add_argument(
        "--relaxation", 
        type=float,
        default=3,
        help='''(float)
             A value between 1 to infinity. 
             This is a parameter that constraints the number of neighbors a cell is allowed to receive.
             Assume dataset 1 has N1 cells, dataset 2 has N2 cells. To find k neighbors in dataset 2 for 
             every cell in dataset 1 means on average each cell in dataset 2 receives (kN1/N2) connections.
             However, not all cells in dataset 2 gets the same number of connections. We therefore set an 
             upper bound for the number of connections a cell in dataset 2 can receive to be:
                (kN1/N2)*relaxation
             where relaxation >= 1. Relaxation=1 enforces a hard limit that every cell receives 
             the same number of nearest neighbors, while relaxation=infinity approaches traditional kNN.
             '''
    )

    advanced.add_argument("--drop_npcs",     type=int, default=0)
    advanced.add_argument(
        "--smoothing_fractions", 
        nargs="+",
        type=float,
        default=[0.7, 0.1, 0.9], # rna, atac, mc
        help='''(list of floats) 
             A list of three values between 0 to 1 that controls the relative contribution
             from the cell itself vs. its neighbors in within-dataset smoothing, 
             specified for 'rna', 'atac', 'mc' data, respectively.
             '''
    )
    advanced.add_argument(
        "--num_pcs", 
        type=int,
        default=50,
        help='''(integer)
             Number of Principal Components to keep for each dataset 
             for smoothing and for clustering/embedding after imputation.
             '''
    )

    return parser

def parse_filename(data_file):
    """turn a xxx/xxx/XXXX.h5ad into XXXX 
    """
    dataset_name = os.path.basename(data_file)
    if dataset_name.endswith('.h5ad'):
        dataset_name = dataset_name[:-len('.h5ad')]
    else:
        raise ValueError("filenames don't have the format xxxx.h5ad")
    return dataset_name

def get_default_mod_direction(mod):
    """
    """
    if mod == 'mc':
        mod_direction = -1
    elif mod == 'rna':
        mod_direction = 1
    elif mod == 'atac':
        mod_direction = 1
    else:
        raise ValueError("choose from ['mc', 'rna', 'atac']")
    return mod_direction 

def get_default_mod_normalization(mod):
    """
    """
    if mod == 'mc':
        mod_norm = 'mc'
    elif mod == 'rna':
        mod_norm = 'cpm'
    elif mod == 'atac':
        mod_norm = 'tpm'
    else:
        raise ValueError("choose from ['mc', 'rna', 'atac']")
    return mod_norm 

def subsampling(mods_selected, settings, metas, gxc_hvftrs, p=0, n=0):
    """Do many datasets at the same time
    p - fraction of cells from each dataset to be included
    """
    if p < 1:
        metas_sub = collections.OrderedDict()
        gxc_hvftrs_sub = collections.OrderedDict()
        for mod in mods_selected: 
            # subsample meta
            if p != 0: 
                cells_included = metas[mod].index.values[np.random.rand(len(metas[mod]))<p]
            elif n != 0:
                if n > len(metas[mod]):
                    n = len(metas[mod])
                    
                cells_included = metas[mod].sample(n=n).index.values
            metas_sub[mod] = metas[mod].loc[cells_included]

            # subsample gxc_hvftrs
            if settings[mod].mod_category == 'mc':
                gxc_hvftrs_sub[mod] = gxc_hvftrs[mod][cells_included]
                logging.info("{} {} {}".format(mod, metas_sub[mod].shape, gxc_hvftrs_sub[mod].shape))
                continue

            cells_included_idx = basic_utils.get_index_from_array(gxc_hvftrs[mod].cell, cells_included)
            gxc_hvftrs_sub[mod] = GC_matrix(
                                            gxc_hvftrs[mod].gene,
                                            cells_included,
                                            gxc_hvftrs[mod].data.tocsc()[:, cells_included_idx],
                                            )
            logging.info("{} {} {}".format(mod, metas_sub[mod].shape, gxc_hvftrs_sub[mod].data.shape))
        return metas_sub, gxc_hvftrs_sub
    else:
        return metas, gxc_hvftrs

def main(args):

    subsample_fraction = args.subsample_fraction
    subsample_times = args.subsample_times

    data_dir = args.data_dir 
    outdir   = args.outdir  
    ka_smooth = args.ka_smooth #
    knn = args.knn

    # # Configs  
    name = args.nametag

    mods_selected = [dtst[:-len('.h5ad')] for dtst in args.input_datasets if dtst.endswith('.h5ad')]
    assert len(mods_selected) == len(args.input_datasets)
    features_selected = [dtst[:-len('.h5ad')] for dtst in args.feature_datasets if dtst.endswith('.h5ad')]
    assert len(features_selected) == len(args.feature_datasets)
    # check features
    for features_modality in features_selected:
        assert (features_modality in mods_selected)
    input_modalities = args.input_modalities
    
    # settings
    Mod_info = collections.namedtuple('Mod_info', [
        'mod', 
        'mod_category',
        'mod_direction', # +1 or -1
    ])
    settings = collections.OrderedDict({})
    for mod, mod_cat in zip(mods_selected, input_modalities):
        settings[mod] = Mod_info(
            mod, 
            mod_cat,
            get_default_mod_direction(mod_cat),
            )

    # within modality
    ps = {mod: frac for mod, frac in zip(['rna', 'atac', 'mc'], args.smoothing_fractions)}
    drop_npcs = {
        'rna':  args.drop_npcs,
        'atac': args.drop_npcs,
        'mc':   args.drop_npcs,
        }

    # across modality
    cross_mod_distance_measure = 'correlation'
    knn = args.knn 
    relaxation = args.relaxation
    n_cca = -1 # 30
    # PCA
    npc = args.num_pcs

    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    # end of configurations

    # ## Read in data 
    logging.info('* Begin integration')

    metas = collections.OrderedDict()
    gxc_hvftrs = collections.OrderedDict()
    for mod in mods_selected:
        # read in the h5ad file
        adata = anndata.read(os.path.join(data_dir, mod+'.h5ad'))
        # metadata (cell level)
        metas[mod] = adata.var.copy() 
        logging.info("Metadata {} {}".format(mod, metas[mod].shape))

        # the feature matrix
        if settings[mod].mod_category == 'mc':
            gxc_hvftrs[mod] = pd.DataFrame(adata.X, index=adata.obs.index.values, columns=adata.var.index.values)
            logging.info("Feature matrix {} {}".format(mod, gxc_hvftrs[mod].shape))
            continue

        gxc_hvftrs[mod] = basic_utils.gc_matrix_from_anndata(adata)
        logging.info("Feature matrix {} {}".format(mod, gxc_hvftrs[mod].data.shape))
    logging.info('Done reading data')

    # subsampling a given fraction of cells from each modality
    for i in range(subsample_times):
        # # specifying output files (one set per subsampling)
        output_pcX_all    = os.path.join(outdir, f'{name}_s{i}_fused_pc'              + f'.h5ad') #  

        # required for downsamp (8/7/2020)
        output_cells      = os.path.join(outdir, f'{name}_s{i}_cells_{{}}'            + f'.txt') # mod 
        # new required arguments (7/27/2020)
        save_knn = True  
        output_knn_within = os.path.join(outdir, f'{name}_s{i}_knn_within_{{}}'       + f'.npz') # mod 
        output_knn_across = os.path.join(outdir, f'{name}_s{i}_knn_across_{{}}_{{}}'  + f'.npz') # modx, mody 
        # # output files end

        logging.info("Subsampling {}/{}".format(i+1, subsample_times))
        metas_sub, gxc_hvftrs_sub = subsampling(
                    mods_selected, settings, metas, gxc_hvftrs, p=subsample_fraction,
                    )
        for mod in mods_selected:
            fout = output_cells.format(mod)
            cells_mod = metas_sub[mod].index.values
            np.savetxt(fout, cells_mod, fmt='%s')

        # ## run SCF
        pcX_all, cells_all = SCF_utils.core_scf_routine(mods_selected, features_selected, settings, 
                                                        metas_sub, gxc_hvftrs_sub, # sub 
                                                        ps, drop_npcs,
                                                        cross_mod_distance_measure, knn, relaxation, n_cca,
                                                        npc,
                                                        output_pcX_all, 
                                                        ka_smooth=ka_smooth,
                                                        save_knn=save_knn,
                                                        output_knn_within=output_knn_within,
                                                        output_knn_across=output_knn_across,
                                                        )
        logging.info('Done integration into a common PC space')
    return 

if __name__ == '__main__':
    log = basic_utils.create_logger()
    parser = argparse.ArgumentParser()
    add_args(parser)
    args = parser.parse_args()
    main(args)



"""For analysis of correlation results
"""
from robustlink._init_plots import *
from robustlink import enhancer_gene_utils

KB = 1000
def p25(x):
    return np.nanpercentile(x, 25)

def p75(x):
    return np.nanpercentile(x, 75)

def estimate_frac_tps(pvalues, bin_n=100, frac_bin_null=0.20):
    """Use the median of the last 5% (frac_bin_null) of bins to estimate null level
    """
    y = np.sort(pvalues)
    bin_edges = np.linspace(0, 1, bin_n)
    bin_width = 1.0/bin_n
    
    bin_counts, bin_edges = np.histogram(y, bin_edges)
    y_norm = bin_counts/(bin_width*bin_counts.sum())
    null_level = np.median(y_norm[-int(frac_bin_null*bin_n):])
    frac_tp = np.clip(1 - null_level, 0, 1)
    
    return frac_tp, null_level, bin_counts, bin_edges, y_norm
    
def plot_pval_dist(ax, frac_tp, null_level, bin_counts, bin_edges, y_norm, fillcolor='C0'):
    """
    """
    _x = bin_edges[:-1]
    ax.plot(_x, y_norm, color='k')
    ax.plot([0, 1], [null_level]*2, linestyle='--', color='k')
    ax.fill_between(_x, null_level, y_norm, 
                    where=y_norm>null_level, alpha=1, color=fillcolor)
    ax.fill_between(_x, 0, np.minimum(y_norm, null_level), alpha=0.5, color='lightgray')
    ax.text(0, null_level+0.2*(1-null_level), 
            "{:.1f}%".format(frac_tp*100), 
            color='white',
            fontsize=18)
    sns.despine(ax=ax)
    return ax

def pipe_plot_pval_dist(ax, pvalues, bin_n=100, frac_bin_null=0.20, fillcolor='C0'):
    """
    """
    # fit distribution 
    frac_tp, null_level, bin_counts, bin_edges, y_norm = estimate_frac_tps(pvalues, bin_n=bin_n, frac_bin_null=frac_bin_null)
    # plot it out 
    plot_pval_dist(ax, 
                   frac_tp, null_level, bin_counts, bin_edges, y_norm, 
                   fillcolor=fillcolor,
                  )
    return 

def quantile_norm(array):
    return pd.Series(array).rank(pct=True, method='average').values

def estimate_frac_tps_vs_dists(dists_kb, res_corrs, res_stats, link_type, FLIP_CORR_SIGN, 
                               bin_n=51, frac_bin_null=0.1,
                              ):
    """Estimate the fraction of true positives with many different distance segments
    """
    frac_tps = []
    num_tps = []
    total_nums = []
    for idx in np.arange(len(dists_kb)):
        dist_kb = dists_kb[idx]
        if idx == 0:
            cond = (res_corrs['dist'] < dist_kb*KB)
        else:
            dist_kb_prev = dists_kb[idx-1]
            cond = ((res_corrs['dist'] < dist_kb*KB) & 
                    (res_corrs['dist'] >= dist_kb_prev*KB))

        if FLIP_CORR_SIGN: 
            corr_sign = -1
        else:
            corr_sign = 1

        pval = np.interp(corr_sign*res_corrs[cond]['corr'], 
                                res_stats['bins'][1:], res_stats[link_type])
        frac_tp, null_level, bin_counts, bin_edges, y_norm = estimate_frac_tps(pval, bin_n=bin_n, frac_bin_null=frac_bin_null)
        frac_tps.append(frac_tp)
        total_nums.append(len(pval))
        num_tps.append(len(pval)*frac_tp)
    return dists_kb, frac_tps, num_tps, total_nums

def estimate_num_sigs_vs_dists(fdr_th, dists_kb, res_corrs, res_stats, link_type, FLIP_CORR_SIGN):
    """Estimate the number of significant pairs with many different distance segments
    """
    frac_tps = []
    num_tps = []
    total_nums = []
    
    sig_gene_set = set()
    cum_sig_genes = []
    
    for idx in np.arange(len(dists_kb)):
        dist_kb = dists_kb[idx]
        if idx == 0:
            cond = (res_corrs['dist'] < dist_kb*KB)
        else:
            dist_kb_prev = dists_kb[idx-1]
            cond = ((res_corrs['dist'] < dist_kb*KB) & 
                    (res_corrs['dist'] >= dist_kb_prev*KB))

        if FLIP_CORR_SIGN: 
            corr_sign = -1
        else:
            corr_sign = 1

        pval = np.interp(corr_sign*res_corrs[cond]['corr'], 
                                res_stats['bins'][1:], res_stats[link_type])
        fdr = pval/quantile_norm(pval)
        
        total_num = len(pval)
        total_nums.append(total_num)
        
        sig_cond = fdr<fdr_th
        sig_genes = res_corrs[cond][sig_cond]['gene']
        sig_gene_set = sig_gene_set.union(set(list(sig_genes.values)))
        cum_sig_gene = len(sig_gene_set)
        cum_sig_genes.append(cum_sig_gene)
        
        num_tp = sig_cond.sum()
        num_tps.append(num_tp)
        
        frac_tp = num_tp/total_num
        frac_tps.append(frac_tp)
        
    return dists_kb, frac_tps, num_tps, total_nums, cum_sig_genes

class CorrRes():
    def __init__(self, res_corrs, pcorr, label="enh-gene", color="C0", lightcolor="lightblue"):
        """
        """
        self.res_corrs = res_corrs
        assert isinstance(pcorr, bool)
        self.pcorr = pcorr # True or False
        self.label = label
        self.color = color
        self.lightcolor = lightcolor
        self.res_stats = None
        
        if pcorr: # positive correlation has a sign of -1 by convention
            self.sign = -1 
        else:
            self.sign = 1
        
    def test_significance(self, fdr=0.2, dist_th=1e5, nbins=501, 
                          output_linked=None, output_correlated=None): # 100kb
        pcorr = self.pcorr
        res_corrs = self.res_corrs
        pval_type_shuffled, pval_type_shuffled_cells = 'left', 'both'
        res_stats = enhancer_gene_utils.get_significance_stats(
                                    res_corrs[['gene', 'enh', 'dist']],
                                    res_corrs['corr'],
                                    res_corrs['corr_shuff'],
                                    res_corrs['corr_shuff_cells'],
                                    pval_type_shuffled, pval_type_shuffled_cells,
                                    bins=np.linspace(-1,1,nbins),
                                    distance_threshold=dist_th,
                                    fdr_threshold=fdr,
                                    positive_side=pcorr,
                                    return_pval=True,
                                    return_cdf=False,
                                )
        if output_linked is not None:
            res_stats['linked_table'].to_csv(output_linked, sep="\t", header=True, index=False)
            print(f"saved to {output_linked}")
        if output_correlated is not None:
            res_stats['correlated_table'].to_csv(output_correlated, sep="\t", header=True, index=False)
            print(f"saved to {output_correlated}")
            
        self.res_stats = res_stats
    
    def estimate_dist_dependence(self, fdr=0.2):
        """Estimate the fraction of true links among all pairs (as a function of distance)
        """
        dists_kb = np.arange(20, 500+1, 20)
        dists_kb_plot = np.hstack([[2], dists_kb])
        
        res_corrs = self.res_corrs
        res_stats = self.res_stats
        if self.pcorr: # atac - positive correlation -- need flipping by convention
            _flip = True
        else:
            _flip = False
        
        for link_type in ['linked_pval', 'correlated_pval']:
            _, frac_tps, num_tps, total_nums = estimate_frac_tps_vs_dists(
                    dists_kb, res_corrs, res_stats, link_type, FLIP_CORR_SIGN=_flip, 
                    bin_n=51, frac_bin_null=0.1,)
            _, sig_frac_tps, sig_num_tps, sig_total_nums, cum_sig_genes = estimate_num_sigs_vs_dists(
                    fdr, 
                    dists_kb, res_corrs, res_stats, link_type, FLIP_CORR_SIGN=_flip)

            # organize results
            mats = np.vstack([
                        dists_kb, 
                        frac_tps, num_tps, total_nums,
                        sig_frac_tps, sig_num_tps, sig_total_nums, cum_sig_genes,
                        ]).T 
            cols = ['dist', 
                    'frac_tp', 'num_pos', 'num_total', 
                    'sig_frac_tp', 'sig_num_pos', 'sig_num_total', 'cum_sig_genes',
                   ] 
            res = pd.DataFrame(mats, columns=cols) 
            if link_type == 'linked_pval': self.distdep_linked = res
            if link_type == 'correlated_pval': self.distdep_correlated = res
            
        return 
        
    def plot_corr_vs_dist(self, ax=None):
        """
        """
        res_corrs = self.res_corrs
        label = self.label
        color = self.color
        if ax is None:
            fig, ax = plt.subplots(figsize=(6,4))
        
        n = 100
        _dists = np.linspace(2*KB, 1000*KB, n)
        res_corrs_median = (res_corrs[['corr']].groupby(pd.cut(res_corrs['dist'], _dists)) 
                                               .agg([np.median, np.mean, p25, p75,])['corr']
                            )

        _x = _dists[1:]
        ax.hlines(0, 0, np.max(_x), linestyle='--', color='k')

        _yme = res_corrs_median['median'].values
        _ylo = res_corrs_median['p25'].values
        _yhi = res_corrs_median['p75'].values
        ax.plot(_x, _yme, linewidth=3, label=label, color=color)
        ax.fill_between(_x, _ylo, _yhi, alpha=0.2, color=color)
        sns.despine(ax=ax)

        ax.set_title('All enhancer-gene pairs\n2kb - 1Mb')
        ax.set_ylabel('Spearman correlation\n(median +/- interquartile)')
        ax.set_xlabel('Enhancer - gene (TSS) distance')
        ax.legend(bbox_to_anchor=(1,0), loc='lower right')
        ax.xaxis.set_major_formatter(mtick.EngFormatter())
        ax.set_ylim([-0.5, 0.5])
        ax.grid(False)
    
    def plot_pval(self, ax=None, dist_th=5e5, bin_n=51, frac_bin_null=0.05, pval_type='linked'):
        """
        """
        label = self.label
        color = self.color
        sign  = self.sign
        res_corrs = self.res_corrs
        res_stats = self.res_stats
        
        if ax is None:
            fig, ax = plt.subplots(figsize=(8,6)) 
        
        if pval_type == 'linked':
            xlabel = 'p (shuff regions as null)'
            stat_col = 'linked_pval'
        elif pval_type == 'correlated':
            xlabel = 'p (shuff metacells as null)'
            stat_col = 'correlated_pval'
        else:
            raise ValueError("choose from linked or correlated")
            
        pvals = np.interp(sign*res_corrs[res_corrs['dist']<dist_th]['corr'], 
                          res_stats['bins'][1:], res_stats[stat_col])
        pipe_plot_pval_dist(ax, pvals, bin_n=bin_n, frac_bin_null=frac_bin_null, fillcolor=color)
        
        ax.set_ylim([0, 2])
        ax.annotate(label, xy=(1,1), xycoords='axes fraction', ha='right', va='top')
        ax.set_xlabel(xlabel)
        ax.set_ylabel('Density')
    
    def plot_dist_dep(self, col, link_type, ax=None):
        """
        col: choose from 
            - frac_tp:     fraction of true positives
            - sig_frac_tp: significant fraction of positives
            - num_pos:     number of positives
            - sig_num_pos: significant number of positives
            - cum_sig_genes: number of significant genes
            
        link_type: choose from
            - linked
            - correlated
        """
        col_labels = {
            'frac_tp': 'Estimated fraction of true links', 
            'sig_frac_tp': 'Fraction of individually\n significant pairs', 
            'num_pos': 'Estimated cumulative\n number of true links',
            'sig_num_pos': 'Cumulative number of\n significant pairs',
            'cum_sig_genes': 'Cumulative number of\n significant genes',
            }
        
        if ax is None:
            fig, ax = plt.subplots(figsize=(5, 4))
            
        df = getattr(self, f"distdep_{link_type}")
        if   link_type == 'linked':     marker = 'o'
        elif link_type == 'correlated': marker = '^'
        
        _x = df['dist'] 
        _y = df[col]
        if col in ['num_pos', 'sig_num_pos']:
            _y = np.cumsum(_y)
        if col in ['num_pos', 'sig_num_pos', 'cum_sig_genes']:
            ax.set_yscale('log')
            ax.yaxis.set_major_formatter(mtick.EngFormatter()) 
        if col in ['frac_tp', 'sig_frac_tp']:
            ax.yaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: '{:.0%}'.format(x))) 
            
        ylabel = col_labels[col]
        ax.plot(_x, _y, f'-{marker}', color=self.color, label=f"{self.label} {link_type}", markersize=7)
        ax.set_xlabel('Enhancer - TSS distance (kb)')
        ax.set_ylabel(ylabel)
        sns.despine(ax=ax)
    
    def plot_corr_distribution(self, ax=None, 
                               bins=np.linspace(-1,1,101),
                              ):
        """
        """
        res_corrs = self.res_corrs
        res_stats = self.res_stats
        sign = self.sign
        if ax is None:
            fig, ax = plt.subplots(figsize=(7,4))
        fontsize = ax.xaxis.label.get_fontsize()
                             
        with sns.axes_style('ticks', {'axes.grid': False}):
            ax.set_title(self.label, pad=20)
            labels = ['shuffled metacells',
                      'shuffled genes',
                      'pairs <500kb',
                      'pairs <100kb',
                      ]
            colors = ['gray', 'black', self.lightcolor, self.color]
            corr_tracks = [
                   res_corrs['corr_shuff_cells'].values,
                   res_corrs['corr_shuff'].values,
                   res_corrs.loc[res_corrs['dist']<=500*KB, 'corr'].values,
                   res_corrs.loc[res_corrs['dist']<=100*KB, 'corr'].values,
                  ]
            vertical_lines = [
                sign*res_stats['r_th_linked'],
                sign*res_stats['r_th_correlated_left'],
                sign*res_stats['r_th_correlated_right'],
            ]

            histy_max = 0
            for j, (_x, label, color) in enumerate(zip(
                    corr_tracks, labels, colors, )):
                # go over columns
                label_comp = '{} ({})'.format(label, len(_x))
                g = ax.hist(_x, bins=bins, histtype='step', label=label, color=color, density=True)
                histy, histx, _ = g
                if j == 0:
                    histy0_max = np.max(histy)
                    histx0_max = bins[np.argmax(histy)]
                elif j > 0:
                    histy_max = max(histy_max, np.max(histy)) 

            ax.set_ylim([0, 1.3*histy_max])
            if histy0_max > 1.3*histy_max:
                # text
                text_config = {
                    'xy': (histx0_max, 1.3*histy_max), 
                    'ha': 'center', 'va': 'bottom', 
                    'xytext': (0, 0),
                    'textcoords': 'offset points',
                    'fontsize': 0.7*fontsize,
                }
                ax.annotate("{:.2f}".format(histy0_max), **text_config)

            # labels
            ax.set_xlim([-1, 1])
            ax.set_xlabel('Spearman correlation')
            ax.set_ylabel('Density')
            sns.despine(ax=ax)

            # line ticks
            ax.grid(which='major', axis='x', linestyle='--')
            ax.xaxis.set_major_formatter(mtick.StrMethodFormatter('{x:.2f}'))
            ax.set_xticks(np.sort(np.hstack([[-1, 0, 1], vertical_lines])))

            # horizontal lines
            lineys = [1.1*histy_max, 1.2*histy_max, 1.2*histy_max]
            linecolors = ['k', 'gray', 'gray']
            texts = ['linked', 'correlated', 'correlated']
            vas = ['top', 'bottom', 'bottom']
            offsets = [(0.2*fontsize, -0.2*fontsize), 
                       (0.2*fontsize, +0.2*fontsize), 
                       (0.2*fontsize, +0.2*fontsize), 
                      ]
            for xcoord, linecolor, liney, text, va, offset in zip(
                vertical_lines, linecolors, lineys, texts, vas, offsets):
                if xcoord < 0:
                    _x = -1
                    xmin, xmax = -1, xcoord
                    ha = 'left'
                else:
                    _x = 1
                    xmin, xmax = xcoord, 1 
                    ha = 'right'
                # line
                ax.hlines(liney, xmin=xmin, xmax=xmax, color=linecolor, linestyle='-')
                # text
                text_config = {
                    'xy': (_x, liney), 
                    'ha': ha, 'va': va, 
                    'xytext': offset,
                    'textcoords': 'offset points',
                    'fontsize': fontsize,
                }
                ax.annotate(text, **text_config)
                
            handles, labels = ax.get_legend_handles_labels()
            handles = [mpl.lines.Line2D([], [], c=h.get_edgecolor()) for h in handles]
            ax.legend(handles, labels, bbox_to_anchor=(1,1), loc='upper left')

    def plot_corr_bimodal(self, othr, dist_th=1e5):
        """
        """
        assert isinstance(othr, CorrRes)

        res1_corrs = self.res_corrs
        res2_corrs = othr.res_corrs
        res1_stats = self.res_stats
        res2_stats = othr.res_stats
        
        assert np.all(res1_corrs.index.values == res2_corrs.index.values)
        
        _x = res1_corrs[res1_corrs['dist']<dist_th]['corr'].values
        _y = res2_corrs[res2_corrs['dist']<dist_th]['corr'].values
        
        r1 = self.sign*res1_stats['r_th_linked']
        r2 = othr.sign*res2_stats['r_th_linked']

        sets = [
            set(res1_stats['linked_table'].index.values),
            set(res2_stats['linked_table'].index.values),
            ]
        num_sig_both = len(sets[0] & sets[1])
        num_sig1 = len(sets[0]) - num_sig_both
        num_sig2 = len(sets[1]) - num_sig_both

        # plot
        fig, ax = plt.subplots(figsize=(6, 6))
        g = ax.hexbin(_x, _y,  
                    gridsize=(100,100),
                    extent=(-1,1,-1,1),
                    cmap='rocket_r', 
                    bins='log', # log10(i+1)
                    rasterized=True,
                    )
        ax.axvline(r1, color='gray', linestyle='--', zorder=2)
        ax.axhline(r2, color='gray', linestyle='--', zorder=2)
        ax.set_aspect('equal')
        ax.set_xlim([-1,1])
        ax.set_ylim([-1,1])  
        ax.set_xticks([-1, r1, 1])
        ax.set_yticks([-1, r2, 1])
        ax.xaxis.set_major_formatter(mtick.FormatStrFormatter("%.2f"))
        ax.yaxis.set_major_formatter(mtick.FormatStrFormatter("%.2f"))
        ax.set_xlabel(f'{self.label} correlation\n(Spearman)')
        ax.set_ylabel(f'{othr.label} correlation\n(Spearman)')
        ax.set_title(f'All enhancer-gene pairs (2-{dist_th/1000:.1f} kb)', pad=0)

        # annotate
        ax.annotate("{} linked\n(mCG-RNA only)".format(num_sig1), 
                    xy=(0,0), xycoords='axes fraction',
                    xytext=(0, -2*ax.xaxis.label.get_fontsize()), textcoords='offset points',
                    ha='right', va='top',
                    fontsize=ax.title.get_fontsize(),
                )
        ax.annotate("{} linked\n(ATAC-RNA only)".format(num_sig2), 
                    xy=(1,1), xycoords='axes fraction',
                    xytext=(0, +ax.title.get_fontsize()), textcoords='offset points',
                    ha='left', va='bottom',
                    fontsize=ax.title.get_fontsize(),
                )
        ax.annotate("{} linked\n(both)".format(num_sig_both), 
                    xy=(0,1), xycoords='axes fraction',
                    xytext=(0, +ax.title.get_fontsize()), textcoords='offset points',
                    ha='right', va='bottom',
                    fontsize=ax.title.get_fontsize(),
                )

        # lines
        ax.plot([-1.2,-0.9], [-1.2, -0.9], color='k', clip_on=False)
        ax.plot([ 1.1, 0.9], [ 1.1,  0.9], color='k', clip_on=False)
        ax.plot([-1.1,-0.9], [ 1.1,  0.9], color='k', clip_on=False)

        # cbar
        cbar = fig.colorbar(g, ax=ax, 
                            fraction=0.05, aspect=10,
                            label='Num. pairs per pixel\n(1/100 length)')
        cbar.ax.yaxis.set_major_formatter(mtick.EngFormatter())
        plt.show()


from setuptools import setup

setup(
    name='robustlink',
    version='0.1.0',
    description='A python package for linking genes and enhancers',
    url='https://github.com/FangmingXie/robustlink',
    author='Fangming Xie, Eran Mukamel',
    author_email='emukamel@ucsd.edu',
    license='MIT',
    packages=['robustlink'],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.8',
    ],
    install_requires=[
        'numpy',
    ]
    )


"""Import commonly used libraries"""

import numpy as np
import pandas as pd
import collections

# matplotlib
import matplotlib as mpl
import matplotlib.pyplot as plt
mpl.rcParams['pdf.fonttype'] = 42 # editable text in matplotlib
mpl.rcParams['svg.fonttype'] = 'none'
# mpl.rcParams['font.family'] = 'sans-serif'
# mpl.rcParams['font.sans-serif'] = ['Arial']

import matplotlib.ticker as mtick
# PercentFormat = mtick.FuncFormatter(lambda y, _: '{:.1%}'.format(y))
# ScalarFormat = mtick.ScalarFormatter()

# seaborn
import seaborn as sns
sns.set_style('ticks', rc={'axes.grid':True})
sns.set_context('talk')

# # set matplotlib formats
# from IPython.display import set_matplotlib_formats
# set_matplotlib_formats('retina')

# empty rectangle (for legend)
EMPTY_RECTANGLE = mpl.patches.Rectangle((0,0), 1, 1, fill=False, edgecolor='none',
		                                 visible=False)

DEFAULT_COLORBAR_KWS = {'fraction': 0.05, 
						'shrink': 0.4, 
						'aspect': 5, 
						}



"""Import commonly used libraries"""

import time
import logging
import glob
import os
import numpy as np
import pandas as pd
import collections
# from natsort import natsorted

# matplotlib
import matplotlib as mpl
import matplotlib.pyplot as plt
mpl.rcParams['pdf.fonttype'] = 42 # editable text in matplotlib
mpl.rcParams['svg.fonttype'] = 'none'

import matplotlib.ticker as mtick
PercentFormat = mtick.FuncFormatter(lambda y, _: '{:.3%}'.format(y))
ScalarFormat = mtick.ScalarFormatter()

# seaborn
import seaborn as sns
sns.set_style('ticks', rc={'axes.grid':True})
sns.set_context('talk')

# # set matplotlib formats
# from IPython.display import set_matplotlib_formats
# set_matplotlib_formats('retina')

# data structures
GC_matrix = collections.namedtuple('GC_matrix', ['gene', 'cell', 'data'])



"""Utility functions for clusterings and embeddings
"""

from .__init__ import *
import igraph as ig
from scipy import sparse
from annoy import AnnoyIndex

from .basic_utils import create_logger

# major change in annoy functions 5/7/2019 
def build_knn_map(X, metric='euclidean', n_trees=10, verbose=True):
    """X is expected to have low feature dimensions (n_obs, n_features) with (n_features <= 50)

    return:
         t: annoy knn object, can be used in the following ways 
                t.get_nns_by_vector
                t.get_nns_by_item
    """
    ti = time.time()

    n_obs, n_f = X.shape
    t = AnnoyIndex(n_f, metric=metric)  # Length of item vector that will be indexed
    for i, X_row in enumerate(X):
        t.add_item(i, X_row)
    t.build(n_trees) # 10 trees
    if verbose:
        print("Time used to build kNN map {}".format(time.time()-ti))
    return t 

def get_knn_by_items(t, k, 
    form='list', 
    search_k=-1, 
    include_distances=False,
    verbose=True, 
    ):
    """Get kNN for each item in the knn map t
    """
    ti = time.time()
    # set up
    n_obs = t.get_n_items()
    n_f = t.f
    if k > n_obs:
        print("Actual k: {}->{} due to low n_obs".format(k, n_obs))
        k = n_obs

    knn = [0]*(n_obs)
    knn_dist = [0]*(n_obs)
    # this block of code can be optimized
    if include_distances:
        for i in range(n_obs):
            res = t.get_nns_by_item(i, k, search_k=search_k, include_distances=include_distances)
            knn[i] = res[0]
            knn_dist[i] = res[1]
    else:
        for i in range(n_obs):
            res = t.get_nns_by_item(i, k, search_k=search_k, include_distances=include_distances) 
            knn[i] = res

    knn = np.array(knn)
    knn_dist = np.array(knn_dist)

    if verbose:
        print("Time used to get kNN {}".format(time.time()-ti))

    if form == 'adj':
        # row col 1/dist 
        row_inds = np.repeat(np.arange(n_obs), k)
        col_inds = np.ravel(knn)
        if include_distances:
            data = np.ravel(knn_dist) 
        else:
            data = [1]*len(row_inds)
        knn_dist_mat = sparse.coo_matrix((data, (row_inds, col_inds)), shape=(n_obs, n_obs))
        return knn_dist_mat
    elif form == 'list':  #
        if include_distances:
            return knn, knn_dist
        else:
            return knn
    else:
        raise ValueError("Choose from 'adj' and 'list'")

def get_knn_by_vectors(t, X, k, 
    form='list', 
    search_k=-1, 
    include_distances=False,
    verbose=True, 
    ):
    """Get kNN for each row vector of X 
    """
    ti = time.time()
    # set up
    n_obs = t.get_n_items()
    n_f = t.f
    n_obs_test, n_f_test = X.shape
    assert n_f_test == n_f

    if k > n_obs:
        print("Actual k: {}->{} due to low n_obs".format(k, n_obs))
        k = n_obs

    knn = [0]*(n_obs_test)
    knn_dist = [0]*(n_obs_test)
    if include_distances:
        for i, vector in enumerate(X):
            res = t.get_nns_by_vector(vector, k, search_k=search_k, include_distances=include_distances) 
            knn[i] = res[0]
            knn_dist[i] = res[1]
    else:
        for i, vector in enumerate(X):
            res = t.get_nns_by_vector(vector, k, search_k=search_k, include_distances=include_distances) 
            knn[i] = res

    knn = np.array(knn)
    knn_dist = np.array(knn_dist)

    if verbose:
        print("Time used to get kNN {}".format(time.time()-ti))

    if form == 'adj':
        # row col 1/dist 
        row_inds = np.repeat(np.arange(n_obs_test), k)
        col_inds = np.ravel(knn)
        if include_distances:
            data = np.ravel(knn_dist) 
        else:
            data = [1]*len(row_inds)
        knn_dist_mat = sparse.coo_matrix((data, (row_inds, col_inds)), shape=(n_obs_test, n_obs))
        return knn_dist_mat
    elif form == 'list':  #
        if include_distances:
            return knn, knn_dist
        else:
            return knn
    else:
        raise ValueError("Choose from 'adj' and 'list'")

def gen_knn_annoy(X, k, form='list', 
    metric='euclidean', n_trees=10, search_k=-1, verbose=True, 
    include_distances=False,
    ):
    """X is expected to have low feature dimensions (n_obs, n_features) with (n_features <= 50)
    """
    ti = time.time()

    n_obs, n_f = X.shape
    t = build_knn_map(X, metric=metric, n_trees=n_trees, verbose=verbose)

    return get_knn_by_items(t, k,                             
                            form=form, 
                            search_k=search_k, 
                            include_distances=include_distances,
                            verbose=verbose, 
                            )

def gen_knn_annoy_train_test(X_train, X_test, k, 
    form='list', 
    metric='euclidean', n_trees=10, search_k=-1, verbose=True, 
    include_distances=False,
    ):
    """X is expected to have low feature dimensions (n_obs, n_features) with (n_features <= 50)
    For each row in X_test, find k nearest neighbors in X_train
    """
    ti = time.time()
    
    n_obs, n_f = X_train.shape
    n_obs_test, n_f_test = X_test.shape
    assert n_f == n_f_test 
    
    t = build_knn_map(X_train, metric=metric, n_trees=n_trees, verbose=verbose)
    return get_knn_by_vectors(t, X_test, k, 
                                form=form, 
                                search_k=search_k, 
                                include_distances=include_distances,
                                verbose=verbose, 
                                )
    
def compute_jaccard_weights_from_knn(X):
    """compute jaccard index on a knn graph
    Arguments: 
        X (unweighted) kNN ajacency matrix (each row Xi* gives the kNNs of cell i) 
        X has to be 0-1 valued 
        k (number of nearest neighbors) 
        
    output: numpy matrix Y
    """
    X = sparse.csr_matrix(X)
    ni, nj = X.shape
    assert ni == nj
    
    k = X[0, :].sum() # number of neighbors
    
    Y = X.dot(X.T)
    # Y = X.multiply(tmp/(2*k - tmp.todense()))    
    Y.data = Y.data/(2*k - Y.data)
    
    return Y 

def adjacency_to_igraph(adj_mtx, weighted=False):
    """
    Converts an adjacency matrix to an igraph object
    
    Args:
        adj_mtx (sparse matrix): Adjacency matrix
        directed (bool): If graph should be directed
    
    Returns:
        G (igraph object): igraph object of adjacency matrix
    
    Uses code from:
        https://github.com/igraph/python-igraph/issues/168
        https://stackoverflow.com/questions/29655111

    Author:
        Wayne Doyle 
        (Fangming Xie modified) 
    """
    nrow, ncol = adj_mtx.shape
    if nrow != ncol:
        raise ValueError('Adjacency matrix should be a square matrix')
    vcount = nrow
    sources, targets = adj_mtx.nonzero()
    edgelist = list(zip(sources.tolist(), targets.tolist()))
    G = ig.Graph(n=vcount, edges=edgelist, directed=True)
    if weighted:
        G.es['weight'] = adj_mtx.data
    return G

def leiden_lite(g, cell_list, resolution=1, weighted=False, verbose=True, num_starts=None, seed=1):
    """ Code from Ethan Armand and Wayne Doyle, ./mukamel_lab/mop
    slightly modified by Fangming Xie 05/13/2019
    """
    import leidenalg
    
    ti = time.time()
    
    if num_starts is not None:
        np.random.seed(seed)
        partitions = []
        quality = []
        seeds = np.random.randint(10*num_starts, size=num_starts)
        for seed in seeds:
            if weighted:
                temp_partition = leidenalg.find_partition(g,
                                                      leidenalg.RBConfigurationVertexPartition, 
                                                      weights=g.es['weight'],
                                                      resolution_parameter=resolution,
                                                      seed=seed,
                                                      )
            else:
                temp_partition = leidenalg.find_partition(g,
                                                      leidenalg.RBConfigurationVertexPartition,
                                                      resolution_parameter=resolution,
                                                      seed=seed,
                                                      )
            quality.append(temp_partition.quality())
            partitions.append(temp_partition)
        partition1 = partitions[np.argmax(quality)]
    else:
        if weighted:
            partition1 = leidenalg.find_partition(g,
                                                  leidenalg.RBConfigurationVertexPartition,
                                                  weights=g.es['weight'],
                                                  resolution_parameter=resolution,
                                                  seed=seed,
                                                  )
        else:
            partition1 = leidenalg.find_partition(g,
                                                  leidenalg.RBConfigurationVertexPartition,
                                                  resolution_parameter=resolution,
                                                  seed=seed,
                                                  )

    # get cluster labels from partition1
    labels = [0]*(len(cell_list)) 
    for i, cluster in enumerate(partition1):
        for element in cluster:
            labels[element] = i+1

    df_res = pd.DataFrame(index=cell_list)
    df_res['cluster'] = labels 
    df_res = df_res.rename_axis('sample', inplace=False)
    
    if verbose:
        print("Time spent on leiden clustering: {}".format(time.time()-ti))
        
    return df_res

def clustering_routine(X, cell_list, k, 
    seed=1, verbose=True,
    resolution=1, metric='euclidean', option='plain', n_trees=10, search_k=-1, num_starts=None):
    """
    X is a (n_obs, n_feature) matrix, n_feature <=50 is recommended
    option: {'plain', 'jaccard', ...}
    """
    assert len(cell_list) == len(X)
    
    if option == 'plain':
        g_knn = gen_knn_annoy(X, k, form='adj', metric=metric, 
                              n_trees=n_trees, search_k=search_k, verbose=verbose)
        G = adjacency_to_igraph(g_knn, weighted=False)
        df_res = leiden_lite(G, cell_list, resolution=resolution, seed=seed, 
                            weighted=False, verbose=verbose, num_starts=num_starts)
        
    elif option == 'jaccard':
        g_knn = gen_knn_annoy(X, k, form='adj', metric=metric, 
                              n_trees=n_trees, search_k=search_k, verbose=verbose)
        gw_knn = compute_jaccard_weights_from_knn(g_knn)
        G = adjacency_to_igraph(gw_knn, weighted=True)
        df_res = leiden_lite(G, cell_list, resolution=resolution, seed=seed, 
                            weighted=True, verbose=verbose, num_starts=num_starts)
    else:
        raise ValueError('Choose from "plain" and "jaccard"')
    
    return df_res

def clustering_routine_multiple_resolutions(X, cell_list, k, 
    seed=1, verbose=True,
    resolutions=[1], metric='euclidean', option='plain', n_trees=10, search_k=-1, num_starts=None):
    """
    X is a (n_obs, n_feature) matrix, n_feature <=50 is recommended
    option: {'plain', 'jaccard', ...}
    """
    assert len(cell_list) == len(X)
    
    res = []
    if option == 'plain':
        g_knn = gen_knn_annoy(X, k, form='adj', metric=metric, 
                              n_trees=n_trees, search_k=search_k, verbose=verbose)
        G = adjacency_to_igraph(g_knn, weighted=False)
        for resolution in resolutions:
            df_res = leiden_lite(G, cell_list, resolution=resolution, seed=seed, 
                                weighted=False, verbose=verbose, num_starts=num_starts)
            df_res = df_res.rename(columns={'cluster': 'cluster_r{}'.format(resolution)})
            res.append(df_res)
        
    elif option == 'jaccard':
        g_knn = gen_knn_annoy(X, k, form='adj', metric=metric, 
                              n_trees=n_trees, search_k=search_k, verbose=verbose)
        gw_knn = compute_jaccard_weights_from_knn(g_knn)
        G = adjacency_to_igraph(gw_knn, weighted=True)
        for resolution in resolutions:
            df_res = leiden_lite(G, cell_list, resolution=resolution, seed=seed, 
                                weighted=True, verbose=verbose, num_starts=num_starts)
            df_res = df_res.rename(columns={'cluster': 'cluster_r{}'.format(resolution)})
            res.append(df_res)
        
    else:
        raise ValueError('Choose from "plain" and "jaccard"')
    res = pd.concat(res, axis=1)
    
    return res

def run_umap_lite(X, cell_list, n_neighbors=15, min_dist=0.1, n_dim=2, 
             random_state=1, output_file=None, **kwargs):
    """run umap on X (n_obs, n_features) 
    """
    # from sklearn.decomposition import PCA
    from umap import UMAP

    ti = time.time()

    logging.info("Running UMAP: {} n_neighbors, {} min_dist , {} dim.\nInput shape: {}"
                        .format(n_neighbors, min_dist, n_dim, X.shape))
    
    umap = UMAP(n_components=n_dim, random_state=random_state, 
                n_neighbors=n_neighbors, min_dist=min_dist, **kwargs)
    ts = umap.fit_transform(X)
 
    if n_dim == 2: 
        df_umap = pd.DataFrame(ts, columns=['tsne_x','tsne_y'])
    elif n_dim == 3:
        df_umap = pd.DataFrame(ts, columns=['tsne_x','tsne_y', 'tsne_z'])

    df_umap['sample'] = cell_list 
    df_umap = df_umap.set_index('sample')
    
    if output_file:
        df_umap.to_csv(output_file, sep="\t", na_rep='NA', header=True, index=True)
        logging.info("Saved coordinates to file. {}".format(output_file))

    tf = time.time()
    logging.info("Done. running time: {} seconds.".format(tf - ti))
    
    return df_umap



"""
"""
from .__init__ import *

import subprocess as sp
import os
from scipy import sparse

def diag_matrix(X, rows=np.array([]), cols=np.array([]), threshold=None):
    """Diagonalize a matrix as much as possible
    """
    di, dj = X.shape
    transposed = 0
    
    if di > dj:
        di, dj = dj, di
        X = X.T.copy()
        rows, cols = cols.copy(), rows.copy()
        transposed = 1
        
    # start (di <= dj)
    new_X = X.copy()
    new_rows = rows.copy() 
    new_cols = cols.copy() 
    if new_rows.size == 0:
        new_rows = np.arange(di)
    if new_cols.size == 0:
        new_cols = np.arange(dj)
        
    # bring the greatest values in the lower right matrix to diagnal position 
    for idx in range(min(di, dj)):

        T = new_X[idx: , idx: ]
        i, j = np.unravel_index(T.argmax(), T.shape) # get the coords of the max element of T
        
        if threshold and T[i, j] < threshold:
            dm = idx # new_X[:dm, :dm] is done (0, 1, ..., dm-1) excluding dm
            break
        else:
            dm = idx+1 # new_X[:dm, :dm] will be done

        # swap row idx, idx+i
        tmp = new_X[idx, :].copy()
        new_X[idx, :] = new_X[idx+i, :].copy() 
        new_X[idx+i, :] = tmp 
        
        tmp = new_rows[idx]
        new_rows[idx] = new_rows[idx+i]
        new_rows[idx+i] = tmp

        # swap col idx, idx+j
        tmp = new_X[:, idx].copy()
        new_X[:, idx] = new_X[:, idx+j].copy() 
        new_X[:, idx+j] = tmp 
        
        tmp = new_cols[idx]
        new_cols[idx] = new_cols[idx+j]
        new_cols[idx+j] = tmp
        
    # 
    if dm == dj:
        pass
    elif dm < dj: # free columns

        col_dict = {}
        sorted_col_idx = np.arange(dm)
        free_col_idx = np.arange(dm, dj)
        linked_rowcol_idx = new_X[:, dm:].argmax(axis=0)
        
        for col in sorted_col_idx:
            col_dict[col] = [col]
        for col, key in zip(free_col_idx, linked_rowcol_idx): 
            if key < dm:
                col_dict[key] = col_dict[key] + [col]
            else:
                col_dict[key] = [col]
                
            
        new_col_order = np.hstack([col_dict[key] for key in sorted(col_dict.keys())])
        
        # update new_X new_cols
        new_X = new_X[:, new_col_order].copy()
        new_cols = new_cols[new_col_order]
    else:
        raise ValueError("Unexpected situation: dm > dj")
    
    if transposed:
        new_X = new_X.T
        new_rows, new_cols = new_cols, new_rows
    return new_X, new_rows, new_cols 

def diag_matrix_rows(X, rows=np.array([]), cols=np.array([]),):
    """Diagonalize a matrix as much as possible by only rearrange rows
    """
    di, dj = X.shape
    
    new_X = X.copy()
    new_rows = rows.copy() 
    new_cols = cols.copy() 
    
    # free to move rows
    row_dict = {}
    free_row_idx = np.arange(di)
    linked_rowcol_idx = new_X.argmax(axis=1) # the column with max value for each row
    
    for row, key in zip(free_row_idx, linked_rowcol_idx): 
        if key in row_dict.keys():
            row_dict[key] = row_dict[key] + [row]
        else:
            row_dict[key] = [row]
            
    new_row_order = np.hstack([row_dict[key] for key in sorted(row_dict.keys())])
    # update new_X new_cols
    new_X = new_X[new_row_order, :].copy()
    new_rows = new_rows[new_row_order]
    
    return new_X, new_rows, new_cols 

def get_grad_colors(n, cmap='copper'):
    """Generate n colors from a given colormap (a matplotlib.cm)
    """
    from matplotlib import cm
    cmap = cm.get_cmap(cmap)
    return [cmap(int(i)) for i in np.linspace(0, 255, n)] 

def logcpm(counts):
    """
    Args:
        - gene-cell matrix
    """
    cov = counts.sum(axis=0)
    logcpm = np.log10(counts.divide(cov, axis=1)*1000000 + 1)
    return logcpm

def logtpm(counts, gene_lengths):
    """
    Args:
        - gene-cell matrix
        - gene_lengths: a series indexed by gene_id
    """
    tpm = counts.divide(gene_lengths.loc[counts.index], axis=0)
    cov = tpm.sum(axis=0)
    logtpm = np.log10((tpm.divide(cov, axis=1))*1000000 + 1)
    return logtpm

def sparse_logcpm(gc_matrix, mode='logcpm', lib_size=[]):
    """
    """
    lib_size = np.array(lib_size)
    if np.size(lib_size) == 0:
        lib_size = gc_matrix.data.sum(axis=0)

    lib_size_inv = sparse.diags(np.ravel(1.0/(1e-7+lib_size)))
    cpm = (gc_matrix.data).dot(lib_size_inv*1e6).tocoo()

    if mode == 'logcpm':
        cpm.data = np.log10(cpm.data + 1)
    elif mode == 'cpm':
        pass

    gc_cpm = GC_matrix(
        gc_matrix.gene, 
        gc_matrix.cell, 
        cpm,
    )
    
    return gc_cpm

def sparse_logtpm(gc_matrix, gene_lengths):
    """
    gene_lengths: array like 
    
    """
    gene_lengths = np.array(gene_lengths)
    gene_length_inv = sparse.diags(np.ravel(1.0/gene_lengths))
    tmp = (gene_length_inv).dot(gc_matrix.data).tocoo()
    lib_size_inv = sparse.diags(np.ravel(1.0/tmp.sum(axis=0)))
    
    logtpm = tmp.dot(lib_size_inv*1e6).tocoo()
    logtpm.data = np.log10(logtpm.data + 1)

    gc_logtpm = GC_matrix(
        gc_matrix.gene, 
        gc_matrix.cell, 
        logtpm,
    )
    
    return gc_logtpm

class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

def create_logger(name='log'):
    """
    args: logger name

    return: a logger object
    """
    logging.basicConfig(
        format='%(asctime)s %(message)s', 
        datefmt='%m/%d/%Y %I:%M:%S %p',
        level=logging.INFO)
    return logging.getLogger(name)

def set_value_by_percentile(this, lo, hi):
    """set `this` below or above percentiles to given values
    this (float)
    lo(float)
    hi(float)
    """
    if this < lo:
        return lo
    elif this > hi:
        return hi
    else:
        return this

def mcc_percentile_norm(mcc, low_p=5, hi_p=95):
    """
    set values above and below specific percentiles to be at the value of percentiles 

    args: mcc, low_p, hi_p  

    return: normalized mcc levels
    """
#   mcc_norm = [np.isnan(mcc) for mcc_i in list(mcc)]
    mcc_norm = np.copy(mcc)
    mcc_norm = mcc_norm[~np.isnan(mcc_norm)]

    lo = np.percentile(mcc_norm, low_p)
    hi = np.percentile(mcc_norm, hi_p)

    mcc_norm = [set_value_by_percentile(mcc_i, lo, hi) for mcc_i in list(mcc)]
    mcc_norm = np.array(mcc_norm)

    return mcc_norm

def plot_tsne_values(df, tx='tsne_x', ty='tsne_y', tc='mCH',
                    low_p=5, hi_p=95,
                    s=2,
                    cbar_label=None,
                    output=None, show=True, close=False, 
                    t_xlim='auto', t_ylim='auto', title=None, figsize=(8,6), **kwargs):
    """
    tSNE plot

    xlim, ylim is set to facilitate displaying glial clusters only

    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    fig, ax = plt.subplots(figsize=figsize)

    im = ax.scatter(df[tx], df[ty], s=s, 
        c=mcc_percentile_norm(df[tc].values, low_p=low_p, hi_p=hi_p), **kwargs)
    if title:
        ax.set_title(title)
    else:
        ax.set_title(tc)
    ax.set_xlabel(tx)
    ax.set_ylabel(ty)
    # ax.set_aspect('auto')


    clb = plt.colorbar(im, ax=ax)
    if cbar_label:
        clb.set_label(cbar_label, rotation=270, labelpad=10)

    if t_xlim == 'auto':
        t_xlim = [np.nanpercentile(df[tx].values, 0.1), np.nanpercentile(df[tx].values, 99.9)]
        t_xlim[0] = t_xlim[0] - 0.1*(t_xlim[1] - t_xlim[0])
        t_xlim[1] = t_xlim[1] + 0.1*(t_xlim[1] - t_xlim[0])
        ax.set_xlim(t_xlim)
    elif t_xlim:
        ax.set_xlim(t_xlim)
    else:
        pass  

    if t_ylim == 'auto':
        t_ylim = [np.nanpercentile(df[ty].values, 0.1), np.nanpercentile(df[ty].values, 99.9)]
        t_ylim[0] = t_ylim[0] - 0.1*(t_ylim[1] - t_ylim[0])
        t_ylim[1] = t_ylim[1] + 0.1*(t_ylim[1] - t_ylim[0])
        ax.set_ylim(t_ylim)
    elif t_ylim:
        ax.set_ylim(t_ylim)
    else:
        pass

    fig.tight_layout()
    if output:
        fig.savefig(output)
        print('Saved to ' + output) 
    if show:
        plt.show()
    if close:
        plt.close(fig)

def get_kwcolors(labels, colors):
    """Generate a dictinary of {label: color} using unique labels and a list of availabel colors
    """
    nc = len(colors)
    nl = len(labels)
    n_repeats = int((nl + nc - 1)/nc)
    colors = list(colors)*n_repeats
    
    kw_colors = {l:c for (l,c) in zip(labels, colors)}
    return kw_colors

def rgb2hex(r,g,b):
    """From rgb (255, 255, 255) to hex
    """
    hex = "#{:02x}{:02x}{:02x}".format(int(r),int(g),int(b))
    return hex

def gen_colors(n, l=0.6, s=0.6, colors=None):
    """Generate compatible and distinct hex colors
    """
    if not colors:
        import colorsys
        hs = np.linspace(0, 1, n, endpoint=False)
        rgbs = [rgb2hex(*(256*np.array(colorsys.hls_to_rgb(h, l, s))))
                 for h in hs]
        return rgbs
    else:
        clrs = [colors[i%len(colors)] for i in range(n)] 
        return clrs 

def myScatter(ax, df, x, y, l, 
              s=20,
              sample_frac=None,
              sample_n=None,
              legend_size=None,
              legend_kws=None,
              grey_label='unlabeled',
              shuffle=True,
              random_state=None,
              legend_mode=0,
              kw_colors=False,
              colors=['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C8', 'C9'], **kwargs):
    """
    take an axis object and make a scatter plot

    - kw_colors is a dictinary {label: color}
    """

    import matplotlib.pyplot as plt
    import seaborn as sns
    df = df.copy()
    # shuffle (and copy) data
    if sample_n:
        df = (df.groupby(l).apply(lambda x: x.sample(min(len(x), sample_n), random_state=random_state))
                            .reset_index(level=0, drop=True)
            )
    if sample_frac:
        df = (df.groupby(l).apply(lambda x: x.sample(frac=sample_frac, random_state=random_state))
                            .reset_index(level=0, drop=True)
            )
    if shuffle:
        df = df.sample(frac=1, random_state=random_state)

    if not kw_colors:
        # add a color column
        inds, catgs = pd.factorize(df[l])
        df['c'] = [colors[i%len(colors)] if catgs[i]!=grey_label else 'grey' 
                    for i in inds]
    else:
        df['c'] = [kw_colors[i] if i!=grey_label else 'grey' for i in df[l]]
    
    # take care of legend
    if legend_mode != -1:
        for ind, row in df.groupby(l).first().iterrows():
            ax.scatter(row[x], row[y], c=row['c'], label=ind, s=s, **kwargs)
        
    if legend_mode == -1:
        pass
    elif legend_mode == 0:
        lgnd = ax.legend()
    elif legend_mode == 1:
        # Shrink current axis's height by 10% on the bottom
        box = ax.get_position()
        ax.set_position([box.x0, box.y0 + box.height * 0.1,
                         box.width, box.height * 0.9])
        lgnd = ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.07),
              ncol=6, fancybox=False, shadow=False) 
    elif legend_mode == 2:
        # Shrink current axis's width by 10% on the bottom
        box = ax.get_position()
        ax.set_position([box.x0 + box.width*0.1, box.y0,
                                 box.width*0.8, box.height])

    if legend_kws:
        lgnd = ax.legend(**legend_kws)

    if legend_mode != -1 and legend_size:
        for handle in lgnd.legendHandles:
            handle._sizes = [legend_size] 

    # backgroud (grey)
    df_grey = df.loc[df['c']=='grey']
    if not df_grey.empty:
        ax.scatter(df_grey[x], 
                   df_grey[y],
                   c=df_grey['c'], s=s, **kwargs)
    # actual plot
    df_tmp = df.loc[df['c']!='grey']
    ax.scatter(df_tmp[x], 
               df_tmp[y],
               c=df_tmp['c'], s=s, **kwargs)
    
    return

def plot_tsne_labels_ax(df, ax, tx='tsne_x', ty='tsne_y', tc='cluster_ID', 
                    sample_frac=None,
                    sample_n=None,
                    legend_size=None,
                    legend_kws=None,
                    grey_label='unlabeled',
                    legend_mode=0,
                    s=1,
                    shuffle=True,
                    random_state=None,
                    t_xlim='auto', t_ylim='auto', title=None, 
                    legend_loc='lower right',
                    colors=['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C8', 'C9'], **kwargs):
    """
    tSNE plot

    xlim, ylim is set to facilitate displaying glial clusters only

    # avoid gray-like 'C7' in colors
    # color orders are arranged for exci-inhi-glia plot 11/1/2017
    """
    import matplotlib.pyplot as plt

    myScatter(ax, df, tx, ty, tc,
             s=s,
             sample_frac=sample_frac,
             sample_n=sample_n,
             legend_size=legend_size,
             legend_kws=legend_kws,
             shuffle=shuffle,
             grey_label=grey_label,
             random_state=random_state, 
             legend_mode=legend_mode, 
             colors=colors, **kwargs)

    if title:
        ax.set_title(title)
    else:
        ax.set_title(tc)
    ax.set_xlabel(tx)
    ax.set_ylabel(ty)
    # ax.set_aspect('auto')

    if t_xlim == 'auto':
        t_xlim = [np.nanpercentile(df[tx].values, 0.1), np.nanpercentile(df[tx].values, 99.9)]
        t_xlim[0] = t_xlim[0] - 0.1*(t_xlim[1] - t_xlim[0])
        t_xlim[1] = t_xlim[1] + 0.1*(t_xlim[1] - t_xlim[0])
        ax.set_xlim(t_xlim)
    elif t_xlim:
        ax.set_xlim(t_xlim)
    else:
        pass  

    if t_ylim == 'auto':
        t_ylim = [np.nanpercentile(df[ty].values, 0.1), np.nanpercentile(df[ty].values, 99.9)]
        t_ylim[0] = t_ylim[0] - 0.1*(t_ylim[1] - t_ylim[0])
        t_ylim[1] = t_ylim[1] + 0.1*(t_ylim[1] - t_ylim[0])
        ax.set_ylim(t_ylim)
    elif t_ylim:
        ax.set_ylim(t_ylim)
    else:
        pass

    return


def plot_tsne_labels(df, tx='tsne_x', ty='tsne_y', tc='cluster_ID', 
                    grey_label='unlabeled',
                    sample_frac=None,
                    sample_n=None,
                    legend_size=None,
                    legend_mode=0,
                    legend_kws=None,
                    s=1,
                    random_state=None,
                    output=None, show=True, close=False, 
                    t_xlim='auto', t_ylim='auto', title=None, figsize=(8,6),
                    legend_loc='lower right',
                    colors=['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C8', 'C9'], **kwargs):
    """
    tSNE plot

    xlim, ylim is set to facilitate displaying glial clusters only

    # avoid gray-like 'C7' in colors
    # color orders are arranged for exci-inhi-glia plot 11/1/2017
    """
    import matplotlib.pyplot as plt
    import seaborn as sns
    fig, ax = plt.subplots(figsize=figsize)

    myScatter(ax, df, tx, ty, tc,
             s=s,
             sample_frac=sample_frac,
             sample_n=sample_n,
             legend_size=legend_size,
             legend_kws=legend_kws,
             grey_label=grey_label,
             random_state=random_state, 
             legend_mode=legend_mode, 
             colors=colors, **kwargs)

    if title:
        ax.set_title(title)
    else:
        ax.set_title(tc)
    ax.set_xlabel(tx)
    ax.set_ylabel(ty)
    # ax.set_aspect('auto')

    if t_xlim == 'auto':
        t_xlim = [np.nanpercentile(df[tx].values, 0.1), np.nanpercentile(df[tx].values, 99.9)]
        t_xlim[0] = t_xlim[0] - 0.1*(t_xlim[1] - t_xlim[0])
        t_xlim[1] = t_xlim[1] + 0.1*(t_xlim[1] - t_xlim[0])
        ax.set_xlim(t_xlim)
    elif t_xlim:
        ax.set_xlim(t_xlim)
    else:
        pass  

    if t_ylim == 'auto':
        t_ylim = [np.nanpercentile(df[ty].values, 0.1), np.nanpercentile(df[ty].values, 99.9)]
        t_ylim[0] = t_ylim[0] - 0.1*(t_ylim[1] - t_ylim[0])
        t_ylim[1] = t_ylim[1] + 0.1*(t_ylim[1] - t_ylim[0])
        ax.set_ylim(t_ylim)
    elif t_ylim:
        ax.set_ylim(t_ylim)
    else:
        pass

    if output:
        fig.savefig(output)
        print('Saved to ' + output) 
    if show:
        plt.show()
    if close:
        plt.close(fig)

def plot_tsne_values_ax(df, ax, tx='tsne_x', ty='tsne_y', tc='mCH',
                    low_p=5, hi_p=95,
                    s=2,
                    cbar=True,
                    cbar_ax=None,
                    cbar_label=None,
                    t_xlim='auto', t_ylim='auto', title=None, **kwargs):
    """
    tSNE plot

    xlim, ylim is set to facilitate displaying glial clusters only

    """
    import matplotlib.pyplot as plt


    im = ax.scatter(df[tx], df[ty], s=s, 
        c=mcc_percentile_norm(df[tc].values, low_p=low_p, hi_p=hi_p), **kwargs)
    if title:
        ax.set_title(title)
    else:
        ax.set_title(tc)
    # ax.set_aspect('auto')
    if cbar:
        if cbar_ax:
            clb = plt.colorbar(im, cax=cbar_ax, shrink=0.4)
        else:
            clb = plt.colorbar(im, cax=ax, shrink=1)
        if cbar_label:
            clb.set_label(cbar_label, rotation=270, labelpad=10)

    if t_xlim == 'auto':
        t_xlim = [np.nanpercentile(df[tx].values, 0.1), np.nanpercentile(df[tx].values, 99.9)]
        t_xlim[0] = t_xlim[0] - 0.1*(t_xlim[1] - t_xlim[0])
        t_xlim[1] = t_xlim[1] + 0.1*(t_xlim[1] - t_xlim[0])
        ax.set_xlim(t_xlim)
    elif t_xlim:
        ax.set_xlim(t_xlim)
    else:
        pass  

    if t_ylim == 'auto':
        t_ylim = [np.nanpercentile(df[ty].values, 0.1), np.nanpercentile(df[ty].values, 99.9)]
        t_ylim[0] = t_ylim[0] - 0.1*(t_ylim[1] - t_ylim[0])
        t_ylim[1] = t_ylim[1] + 0.1*(t_ylim[1] - t_ylim[0])
        ax.set_ylim(t_ylim)
    elif t_ylim:
        ax.set_ylim(t_ylim)
    else:
        pass

    return im  


def get_mcc(df, base_call_cutoff=100, sufficient_coverage_fraction=1, suffix=True, fillna=True):
    """Get mcc matrix from mc_c matrix (filtering out low coverage gene or bins)
    """
    logging.info('Getting mcc matrix from mc and c') 
    logging.info('base_call_cutoff={}, sufficient_coverage_fraction={}'.format(
                base_call_cutoff, sufficient_coverage_fraction))
    
    df_c = df.filter(regex="_c$")
    df_c.columns = [col[:-len('_c')] for col in df_c.columns] 
    df_mc = df.filter(regex="_mc$")
    df_mc.columns = [col[:-len('_mc')] for col in df_mc.columns] 
    # a gene is sufficiently covered in % of cells 
    condition = (df_c > base_call_cutoff).sum(axis=1) >= sufficient_coverage_fraction*(df.shape[1])/2.0 

    logging.info("Matrix size before pruning... "+ str(df.shape))
    logging.info("Matrix size after pruning... "+ str(df.loc[condition].shape))
    
    # get mcc matrix with kept bins and nan values for low coverage sites
    df_c_nan = df_c.copy()
    df_c_nan[df_c < base_call_cutoff] = np.nan
    df_mcc = df_mc.loc[condition]/df_c_nan.loc[condition]
    logging.info(df_mcc.shape)

    # imputation (missing value -> mean value of all cells)
    if fillna:
        logging.info('Imputing data... (No effect if sufficient_coverage_fraction=1)')
        means = df_mcc.mean(axis=1)
        fill_value = pd.DataFrame({col: means for col in df_mcc.columns})
        df_mcc.fillna(fill_value, inplace=True)
    
    # add suffix
    if suffix:
        df_mcc.columns = df_mcc.columns.values + '_mcc'
    
    return df_mcc

def get_mcc_lite(mc_table, c_table, base_call_cutoff=100, sufficient_coverage_fraction=1, fillna=True):
    """Given 2 numpy array, return mcc table
    Gene/region by sample matrix
    """
    df_c = pd.DataFrame(c_table)
    df_mc = pd.DataFrame(mc_table)
    assert df_c.shape == df_mc.shape
    
    # a gene is sufficiently covered in % of cells 
    condition = (df_c > base_call_cutoff).sum(axis=1) >= sufficient_coverage_fraction*(df_c.shape[1])

    logging.info("Matrix size before pruning... "+ str(df_c.shape))
    logging.info("Matrix size after pruning... "+ str(df_c.loc[condition].shape))
    
    # get mcc matrix with kept bins and nan values for low coverage sites
    df_c_nan = df_c.copy()
    df_c_nan[df_c < base_call_cutoff] = np.nan
    df_mcc = df_mc.loc[condition]/df_c_nan.loc[condition]
    logging.info(df_mcc.shape)

    # imputation (missing value -> mean value of all cells)
    if fillna:
        logging.info('Imputing data... (No effect if sufficient_coverage_fraction=1)')
        means = df_mcc.mean(axis=1)
        fill_value = pd.DataFrame({col: means for col in df_mcc.columns})
        df_mcc.fillna(fill_value, inplace=True)
    
    # return matrix and index (regions)
    return df_mcc.values, df_mcc.index.values

def get_mcc_lite_v2(df_c, df_mc, base_call_cutoff):
    """
    """
    # get mcc matrix with kept bins and nan values for low coverage sites
    df_c_nan = df_c.copy()
    df_c_nan[df_c < base_call_cutoff] = np.nan
    df_mcc = df_mc/df_c_nan
    logging.info(df_mcc.shape)

    # imputation (missing value -> mean value of all cells)
    means = df_mcc.mean(axis=1)
    fill_value = pd.DataFrame({col: means for col in df_mcc.columns})
    df_mcc.fillna(fill_value, inplace=True)
    
    return df_mcc

def get_mcc_lite_v3(df_c, df_mc, base_call_cutoff):
    """
    """
    # get mcc matrix with kept bins and nan values for low coverage sites
    df_c_nan = df_c.copy()
    df_c_nan[df_c < base_call_cutoff] = np.nan
    df_mcc = df_mc/df_c_nan
    return df_mcc


def get_clusters_mc_c_worker(df_cells, df_input, cluster_col):
    """reduce gene*cell or bin*cell matrix to a gene*cluster or bin*cluster matrix
    Arguments:
        - df_cells: a dataframe indexed by 'cell_name', and have '$cluster_col' as column
        - df_input: a dataframe with 'sample_mc', 'sample_c' ... as columns
        sample names are cell names
    """
    # cluster mc_c
    df_c = df_input.filter(regex='_c$')
    df_mc = df_input.filter(regex='_mc$')

    df_mc_c = pd.DataFrame() 
    for label, df_sub in df_cells.groupby(cluster_col):
        samples = df_sub.index.values
        df_mc_c['{}_mc'.format(label)] = df_mc[samples+'_mc'].sum(axis=1)
        df_mc_c['{}_c'.format(label)] = df_c[samples+'_c'].sum(axis=1)

    logging.info("Output shape: {}".format(df_mc_c.shape))
    return df_mc_c

def rank_array(array):
    """Return ranking of each element of an array
    """
    array = np.array(array)
    temp = array.argsort()
    ranks = np.empty_like(temp)
    ranks[temp] = np.arange(len(array))
    return ranks

# added 4/5/2019
def rank_rows(matrix):
    """Return rankings of each rwo in a 2d array
    """
    matrix = np.array(matrix)
    return np.apply_along_axis(rank_array, 1, matrix) # row = 1

def spearman_corrcoef(X, Y):
    """return spearman correlation matrix for each pair of rows of X and Y
    """
    return np.corrcoef(rank_rows(X), rank_rows(Y))

def spearmanr_paired_rows(X, Y):
    from scipy import stats
    
    X = np.array(X)
    Y = np.array(Y)
    corrs = []
    ps = []
    for x, y in zip(X, Y):
        r, p = stats.spearmanr(x, y)
        corrs.append(r)
    return np.array(corrs), np.array(ps)

def get_index_from_array(arr, inqs, na_rep=-1):
    """Get index of array
    """
    arr = np.array(arr)
    arr = pd.Series(arr).reset_index().set_index(0)
    idxs = arr.reindex(inqs)['index'].fillna(na_rep).astype(int).values
    return idxs

def get_genomic_distance(sa, ea, sb, eb):
    """Get genomic distance
    """
    assert sa < ea and sb < eb
    if sa > sb:
        sa, sb = sb, sa
        ea, eb = eb, ea
        
    # sa <= sb
    distance = max(0, sb - ea)
    
    return distance

def get_reverse_comp(string):
    """Get reverse compliment of a string
    """
    comp_dict = {
        'A': 'T',
        'T': 'A',
        'G': 'C',
        'C': 'G',
        'N': 'N',
    }
    for char in set(string):
        if char not in ['A', 'C', 'G', 'T', 'N']:
            raise ValueError('Not allowed char in string')
            
    new_string = ''.join([comp_dict[char] for char in string[::-1]])
    return new_string
    
# added 4/11/2019
def save_gxc_matrix(gxc, f_mat, f_gene, f_cell):
    """
    """
    print("Deprecated: 6/29/2019 Use save_gc_matrix instead")
    sparse.save_npz(f_mat, gxc.data)
    with open(f_gene, 'w') as f:
        f.write('\n'.join(gxc.gene)+'\n')
    with open(f_cell, 'w') as f:
        f.write('\n'.join(gxc.cell)+'\n')
        
def save_gxc_matrix_methylation(gxc, f_mat_c, f_mat_mc, f_gene, f_cell):
    """
    """
    print("Deprecated: 6/29/2019 Use save_gc_matrix_methylation instead")
    sparse.save_npz(f_mat_c, gxc.data['c'])
    sparse.save_npz(f_mat_mc, gxc.data['mc'])
    with open(f_gene, 'w') as f:
        f.write('\n'.join(gxc.gene)+'\n')
    with open(f_cell, 'w') as f:
        f.write('\n'.join(gxc.cell)+'\n') 

def save_gc_matrix(gc_matrix, f_gene, f_cell, f_mat):
    """
    """
    sparse.save_npz(f_mat, gc_matrix.data)
    with open(f_gene, 'w') as f:
        f.write('\n'.join(gc_matrix.gene)+'\n')
    with open(f_cell, 'w') as f:
        f.write('\n'.join(gc_matrix.cell)+'\n')

def save_gc_matrix_methylation(gc_matrix, f_gene, f_cell, f_mat_mc, f_mat_c):
    """
    """
    sparse.save_npz(f_mat_mc, gc_matrix.data['mc'])
    sparse.save_npz(f_mat_c, gc_matrix.data['c'])
    with open(f_gene, 'w') as f:
        f.write('\n'.join(gc_matrix.gene)+'\n')
    with open(f_cell, 'w') as f:
        f.write('\n'.join(gc_matrix.cell)+'\n') 

def import_single_textcol(fname, header=None, col=0):
    return pd.read_csv(fname, header=header, sep='\t')[col].values

def export_single_textcol(fname, array):
    with open(fname, 'w') as f:
        f.write('\n'.join(array)+'\n')

def load_gc_matrix(f_gene, f_cell, f_mat):
    """
    """
    gene = import_single_textcol(f_gene)
    cell = import_single_textcol(f_cell)
    mat = sparse.load_npz(f_mat) 
    assert (len(gene), len(cell)) == mat.shape
    return GC_matrix(gene, cell, mat) 

def load_gc_matrix_methylation(f_gene, f_cell, f_mat_mc, f_mat_c):
    """
    """
    _gene = import_single_textcol(f_gene) 
    _cell = import_single_textcol(f_cell)
    _mat_mc = sparse.load_npz(f_mat_mc) 
    _mat_c = sparse.load_npz(f_mat_c) 
    gxc_raw = GC_matrix(_gene, _cell, 
                              {'c': _mat_c, 'mc': _mat_mc})
    return gxc_raw

def gc_matrix_from_anndata(adata):
    """assuming a gene by cell AnnData
    """
    gene = adata.obs.index.values
    cell = adata.var.index.values
    mat = adata.X  
    return GC_matrix(gene, cell, mat) 

def nondup_legends(ax='', **kwargs):
    """Assuming plt (matplotlib.pyplot) is imported
    """
    from collections import OrderedDict
    import matplotlib.pyplot as plt

    if ax == '':
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        plt.legend(by_label.values(), by_label.keys(), **kwargs)
    else:
        handles, labels = ax.get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        ax.legend(by_label.values(), by_label.keys(), **kwargs)
    return 

def dedup_array_elements(x, empty_string=''):
    """Replacing repeats with empty_string
    """
    newx = np.empty_like(x)
    newx[0] = x[0]
    for i in range(1, len(x)):
        if x[i-1] == x[i]:
            newx[i] = empty_string
        else:
            newx[i] = x[i]
    return newx

def vcorrcoef(X,Y):
    """Compute correlation coef for each rows of X and Y
    """
    assert X.shape == Y.shape
    Xm = np.mean(X,axis=1).reshape(-1,1)
    Ym = np.mean(Y,axis=1).reshape(-1,1)
    Xm = X-Xm
    Ym = Y-Ym
    
    r_num = np.sum(Xm*Ym,axis=1)
    r_den = np.sqrt(np.sum(Xm**2,axis=1)*np.sum(Ym**2, axis=1))
    r = r_num/r_den
    return r

def zscore(x, offset=1e-7, ddof=1):
    return (x - np.mean(x))/(np.std(x, ddof=ddof) + offset)


def clst_umap_pipe_lite(pcs, cells_all, 
                        resolution=1,
                        npc=50,
                        k=30,
                        verbose=False, seed=0, cluster_only=False, 
                       ):
    # clustering
    import CEMBA_clst_utils
    import CEMBA_run_tsne

    df_clst = CEMBA_clst_utils.clustering_routine(
                                    pcs, 
                                    cells_all, k, 
                                    verbose=verbose,
                                    resolution=resolution,
                                    seed=seed,
                                    metric='euclidean', option='plain', n_trees=10, search_k=-1)

    # umap
    if not cluster_only:
        df_tsne = CEMBA_run_tsne.run_umap_lite(
                    pcs, 
                    cells_all, 
                    verbose=verbose,
                    n_neighbors=30, min_dist=0.5, n_dim=2, 
                    random_state=1)

        df_summary = df_clst.join(df_tsne)
        return df_summary
    else:
        return df_clst

def gen_cdf(array, ax, x_range=[], n_points=1000, show=True, flip=False, **kwargs):
    """
    """
    x = np.sort(array)
    y = np.arange(len(array))/len(array)
    if flip:
        # x = x[::-1]
        y = 1 - y

    if not x_range:
        if show:
            ax.plot(x, y, **kwargs)
        return x, y 
    else:
        start, end = x_range
        xbins = np.linspace(start, end, n_points)
        ybins = np.interp(xbins, x, y)
        if show:
            ax.plot(xbins, ybins, **kwargs)
        return xbins, ybins 

def savefig(fig, path):
    """
    """
    fig.savefig(path, bbox_inches='tight', dpi=300)
    return 


#!/usr/bin/env python
# coding: utf-8

import argparse
import os

import numpy as np
import logging
import fbpca
import itertools
import argparse
import anndata

from robustlink import utils
from robustlink.scf import clst_utils
from robustlink.scf import basic_utils

def pipe_singlemod_clustering(
        f_ann,
        flist_selected_cells, flist_res,
        resolutions,
        npc=50,
        kforleiden=30,
    ):
    """
    """
    # read input
    adata = anndata.read(f_ann) 
    gc_mat = basic_utils.gc_matrix_from_anndata(adata)

    for f_selected_cells, f_res in zip(flist_selected_cells, flist_res):
        print("processing {}".format(f_selected_cells))
        selected_cells = np.loadtxt(f_selected_cells, dtype='str')
        # trim the matrix
        cg_mat_dense = gc_mat.data.T.todense()
        cg_mat_dense = cg_mat_dense[utils.get_index_from_array(gc_mat.cell, selected_cells)]

        # Louvain clustering for different resolutions
        # X should be selected from highly variable genes and normalized
        U, s, Vt = fbpca.pca(cg_mat_dense, k=npc)
        pcX = U.dot(np.diag(s))

        res_clsts = clst_utils.clustering_routine_multiple_resolutions(
                            pcX, selected_cells, kforleiden, 
                            seed=1, verbose=True,
                            resolutions=resolutions, metric='euclidean', option='plain', 
                            n_trees=10, search_k=-1, num_starts=None
                            )

        # organize and save results
        print(f_res)
        res_clsts.to_csv(f_res, sep='\t', na_rep='NA', header=True, index=True)
    return 

def wrapper_singlemod_clustering(
        f_ann,
        out_dir, 
        input_name_tag,
        subsample_times, 
        resolutions=[1],
    ):
    """
    """
    # input data
    shortname = os.path.basename(f_ann)
    assert shortname.endswith('.h5ad')
    shortname = shortname[:-len('.h5ad')]

    # input cell lists
    flist_selected_cells = [os.path.join(out_dir, f'{input_name_tag}_s{i_sub}_cells_{shortname}.txt')
            for i_sub in np.arange(subsample_times)
        ]
    # output files
    flist_res =            [os.path.join(out_dir, f'{input_name_tag}_s{i_sub}_metacells_{shortname}.tsv.gz')
            for i_sub in np.arange(subsample_times)
        ]

    npc = 50
    kforleiden = 30
    pipe_singlemod_clustering(
        f_ann,
        flist_selected_cells, flist_res,
        resolutions,
        npc=npc,
        kforleiden=kforleiden,
    )
    return 

def add_args(parser):
    """
    """
    parser.add_argument("-i", "--input_dataset", help="input dataset (rna); h5ad file", required=True)
    parser.add_argument("-o", "--out_dir", help="output result directory", required=True)
    parser.add_argument("-tag", "--input_name_tag", help="input name tag from scfusion", required=True)
    parser.add_argument("-sn", "--subsample_times", help=">1", type=int, required=True)
    parser.add_argument("-r", "--resolutions", nargs='+', help="Leiden clustering resolutions", required=True)
    return 

def main(args):
    """
    """
    # output setting
    # run this with each combination of (i_sub, knn)
    dataset = args.input_dataset
    assert dataset.endswith('.h5ad')
    out_dir = args.out_dir
    input_name_tag = args.input_name_tag
    subsample_times = args.subsample_times
    resolutions = args.resolutions
    if isinstance(resolutions, str):
        resolutions = [int(resolutions)]
    elif isinstance(resolutions, list):
        resolutions = [int(r) for r in resolutions]

    wrapper_singlemod_clustering(
        dataset,
        out_dir, 
        input_name_tag,
        subsample_times, 
        resolutions=resolutions,
    )

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_args(parser)
    args = parser.parse_args()
    main(args)



"""
"""
import logging
import numpy as np
import pandas as pd
import os

def logcpm(counts):
    """
    Args:
        - gene-cell matrix (pandas DataFrame)
    """
    cov = counts.sum(axis=0)
    logcpm = np.log10(counts.divide(cov, axis=1)*1000000 + 1)
    return logcpm

def logtpm(counts, gene_lengths):
    """
    Args:
        - gene-cell matrix (pandas DataFrame)
        - gene_lengths: a series indexed by gene_id
    """
    tpm = counts.divide(gene_lengths.loc[counts.index], axis=0)
    cov = tpm.sum(axis=0)
    logtpm = np.log10((tpm.divide(cov, axis=1))*1000000 + 1)
    return logtpm

class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

def create_logger(name='log'):
    """
    args: logger name
    return: a logger object
    """
    logging.basicConfig(
        format='%(asctime)s %(message)s', 
        datefmt='%m/%d/%Y %I:%M:%S %p',
        level=logging.INFO)
    return logging.getLogger(name)

def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

def get_mcc_lite_v4(df_c, df_mc, base_call_cutoff, sufficient_coverage_fraction=1, fillna=True):
    """
    """
    # a gene is sufficiently covered in % of cells 
    condition = (df_c > base_call_cutoff).sum(axis=1) >= sufficient_coverage_fraction*(df_c.shape[1]) 

    # get mcc matrix with kept bins and nan values for low coverage sites
    df_c_nan = df_c.copy()
    df_c_nan[df_c < base_call_cutoff] = np.nan
    df_mcc = df_mc.loc[condition]/df_c_nan.loc[condition]

    # imputation (missing value -> mean value of all cells)
    if fillna:
        logging.info('Imputing data... (No effect if sufficient_coverage_fraction=1)')
        means = df_mcc.mean(axis=1)
        fill_value = pd.DataFrame({col: means for col in df_mcc.columns})
        df_mcc.fillna(fill_value, inplace=True)
    
    return df_mcc

def get_index_from_array(arr, inqs, na_rep=-1):
    """Get index of array
    """
    arr = np.array(arr)
    arr = pd.Series(arr).reset_index().set_index(0)
    idxs = arr.reindex(inqs)['index'].fillna(na_rep).astype(int).values
    return idxs

def import_single_textcol(fname, header=None, col=0):
    return pd.read_csv(fname, header=header, sep='\t')[col].values

def export_single_textcol(fname, array):
    with open(fname, 'w') as f:
        f.write('\n'.join(array)+'\n')

def combine_legends(axs_flat):
    """Combine legends from subplots (axs.flat)
    """
    handles_all = [] 
    labels_all = []
    for ax in axs_flat:
        handles, labels = ax.get_legend_handles_labels()
        handles_all.append(handles)
        labels_all.append(labels)
    
    return np.hstack(handles_all), np.hstack(labels_all)



"""Utilities for SingleCellFusion
"""
from .__init__ import *

import functools
import collections
import itertools
import re
from scipy import sparse
from scipy.stats import zscore
import fbpca
import sys
import logging
import fbpca
import anndata

from . import basic_utils
from . import clst_utils

def sparse_adj_to_mat(adjs, row_size, col_size, dists=''):
    """Turn a knn adjacency matrix to a sparse matrix
    """
    n_obs, k = adjs.shape
    assert n_obs == row_size
    # row col 1/dist 
    row_inds = np.repeat(np.arange(row_size), k)
    col_inds = np.ravel(adjs)
    if isinstance(dists, np.ndarray):
        assert dists.shape == adjs.shape
        data = np.ravel(dists) 
    else:
        data = [1]*len(row_inds)
    knn_dist_mat = sparse.coo_matrix((data, (row_inds, col_inds)), shape=(row_size, col_size))
    return knn_dist_mat

# smooth-within modality
def smooth_in_modality(counts_matrix, norm_counts_matrix, k, ka, npc=100, sigma=1.0, p=0.1, drop_npc=0, 
    output_filename="",
    ):
    """Smooth a data matrix
    
    Arguments:
        - counts_matrix (pandas dataframe, feature by cell)
        - norm_counts_matrix (pandas dataframe, feature by cell) log10(CPM+1)
        - k (number of nearest neighbors)
        - ka normalized by the ka'th neighbor (including self)
    Return:
        - smoothed cells_matrix (pandas dataframe)
        - markov affinity matrix
    """
    # from sklearn.neighbors import NearestNeighbors
    # import clst_utils
    
    assert counts_matrix.shape[1] == norm_counts_matrix.shape[1] 

    c = norm_counts_matrix.columns.values
    N = len(c)

    # reduce dimension fast version
    U, s, Vt = fbpca.pca(norm_counts_matrix.T.values, k=npc)
    pcs = U.dot(np.diag(s))
    if drop_npc != 0:
        pcs = pcs[:, drop_npc:]

    # get k nearest neighbor distances fast version 
    inds, dists = clst_utils.gen_knn_annoy(pcs, k, form='list', 
                                                metric='euclidean', n_trees=10, search_k=-1, verbose=True, 
                                                include_distances=True)
    
    # remove itself
    dists = dists[:, 1:]
    inds = inds[:, 1:]

    # normalize by ka's distance 
    # (including itself; needs to -2; one to switch to 0-based index, one account for the remove itself step) 
    dists = (dists/(dists[:, ka-2].reshape(-1, 1)))

    # gaussian kernel
    adjs = np.exp(-((dists**2)/(sigma**2))) 

    # construct a sparse matrix 
    cols = np.ravel(inds)
    rows = np.repeat(np.arange(N), k-1) # remove itself
    vals = np.ravel(adjs)
    A = sparse.csr_matrix((vals, (rows, cols)), shape=(N, N))

    # Symmetrize A (union of connection)
    A = A + A.T

    # normalization fast (A is now a weight matrix excluding itself)
    degrees = A.sum(axis=1)
    A = sparse.diags(1.0/np.ravel(degrees)).dot(A)

    # include itself
    eye = sparse.identity(N)
    A = p*eye + (1-p)*A
    
    # smooth fast (future?)
    counts_matrix_smoothed = pd.DataFrame((A.dot(counts_matrix.T)).T, 
                                         columns=counts_matrix.columns, index=counts_matrix.index)
    if output_filename:
        # save A (name, axis?)
        sparse.save_npz(output_filename, A)
    return counts_matrix_smoothed, A

# impute across modality
def get_constrained_knn(mat_norm_j, mat_norm_i, knn, k_saturate, knn_speed_factor=10, metric='dot', verbose=False, 
    ):
    """Get constrained knn
    j <- i
    Look for kNN in i for each cell in j, cells in i are constrained to k_saturated
    
    get knn_speed_factor*knn number of nearest neighbors internally
    """
    ti = time.time()
    assert mat_norm_i.shape[1] == mat_norm_j.shape[1]
    knn = min(int(knn), len(mat_norm_i)) # min added 08/11/2020
    knn_speed_factor = int(knn_speed_factor)
    
    cells_i = np.arange(len(mat_norm_i))
    cells_j = np.arange(len(mat_norm_j))
    
    # record cells in j
    accepted_knn_ji = [] 
    accepted_cells = []
    rejected_cells = np.arange(len(cells_j))
    
    # record cell in i
    n_connects = np.zeros(len(cells_i)).astype(int) # record number of connection for each cell in i 
    unsaturated = (n_connects < k_saturate) # unsaturated bool 
    unsaturated_cells = np.arange(len(cells_i))[unsaturated]
    
    while rejected_cells.size != 0:
        if verbose:
            logging.info("**{}, {}, {}".format(len(rejected_cells), len(unsaturated_cells), time.time()-ti))
        
        np.random.shuffle(rejected_cells) # random order
        # do something to rejected cells and unsaturated cells
        # knn_ji # for each cell in j, its knn in i
        knn_ji = clst_utils.gen_knn_annoy_train_test(mat_norm_i.values[unsaturated_cells], # look for nearest neighbors in i 
                                                           mat_norm_j.values[rejected_cells], # for each row in j
                                                           min(knn*knn_speed_factor, len(unsaturated_cells)), #  
                                                           form='list', # adj matrix 
                                                           metric=metric, # correlation 
                                                           n_trees=10, search_k=-1, verbose=False, 
                                                           include_distances=False, # for now
                                                           ).astype(int)
        knn_ji = unsaturated_cells[knn_ji] # transform it to global index, need to check this like 
        
        rejected_local_idx = []
        # examine each cell in j
        for local_idx, cell in enumerate(rejected_cells):
            # get knn in i
            knn_in_i = knn_ji[local_idx]
            # filter out saturated ones
            knn_in_i = knn_in_i[unsaturated[knn_in_i]]
            
            if knn_in_i.size < knn:
                # reject
                rejected_local_idx.append(local_idx)
            else:
                # accept and update
                accepted_knn_ji.append(knn_in_i[:knn])
                accepted_cells.append(cell)
                n_connects[knn_in_i[:knn]] += 1 
                unsaturated = (n_connects < k_saturate) # unsaturated bool 
                
        unsaturated_cells = np.arange(len(cells_i))[unsaturated]
        rejected_cells = rejected_cells[rejected_local_idx]
    # break
                
    accepted_knn_ji = pd.DataFrame(np.vstack(accepted_knn_ji), index=accepted_cells)
    accepted_knn_ji = accepted_knn_ji.sort_index().values

    
    return accepted_knn_ji
# 
def impute_1pair_cca(mod_i, mod_j, 
                     smoothed_features_i, smoothed_features_j,
                     settings,
                     knn,
                     relaxation,
                     n_cca,
                     output_knn_mat_ij='',
                     output_knn_mat_ji='',
                     impute_j=True,
                    ):
    """
    """
    # set up
    direct_i, direct_j = settings[mod_i].mod_direction, settings[mod_j].mod_direction
    
    mat_ii = smoothed_features_i.T # cell in mod i; gene in mod i
    mat_jj = smoothed_features_j.T # cell in mod j; gene in mod j
    
    genes_i = mat_ii.columns.values
    genes_j = mat_jj.columns.values
    genes_common = np.intersect1d(genes_i, genes_j)
    
    cells_i = mat_ii.index.values
    cells_j = mat_jj.index.values
    
    ## CCA euclidean distance 
    # normalize the feature matrix
    X = mat_ii[genes_common].T.apply(basic_utils.zscore, axis=0)*direct_i # gene by cell, zscore across genes
    Y = mat_jj[genes_common].T.apply(basic_utils.zscore, axis=0)*direct_j
    U, s, Vt = fbpca.pca(X.T.values.dot(Y.values), k=n_cca)
    del X, Y

    mat_norm_i = pd.DataFrame(U, index=mat_ii.index)
    maxk_i = int((len(cells_j)/len(cells_i))*knn*relaxation)+1 # max number of NN a cell in i can get 
    mat_norm_j = pd.DataFrame(Vt.T, index=mat_jj.index)
    maxk_j = int((len(cells_i)/len(cells_j))*knn*relaxation)+1 # max number of NN a cell in j can get 
    
    if impute_j:
        # knn_i and knn_j
        # j <- i for each j, get kNN in i
        knn_ji = get_constrained_knn(mat_norm_j, mat_norm_i, knn=knn, k_saturate=maxk_i, metric='euclidean')
        mat_knn_ji = sparse_adj_to_mat(knn_ji, len(cells_j), len(cells_i))
        
        if output_knn_mat_ji:
            sparse.save_npz(output_knn_mat_ji, mat_knn_ji)
        
        # normalize 
        degrees_j = np.ravel(mat_knn_ji.sum(axis=1)) # for each cell in j, how many cells in i it connects to 
        mat_knn_ji = sparse.diags(1.0/(degrees_j+1e-7)).dot(mat_knn_ji) 
        
        # imputation both across and within modality
        mat_ji = mat_knn_ji.dot(mat_ii) # cell in mod j, gene in mod i
    
    
    # i <- j
    knn_ij = get_constrained_knn(mat_norm_i, mat_norm_j, knn=knn, k_saturate=maxk_j, metric='euclidean')
    mat_knn_ij = sparse_adj_to_mat(knn_ij, len(cells_i), len(cells_j))
    
    if output_knn_mat_ij:
        sparse.save_npz(output_knn_mat_ij, mat_knn_ij)
    
    degrees_i = np.ravel(mat_knn_ij.sum(axis=1)) # for each cell in i, how many cells in j it connects to 
    mat_knn_ij = sparse.diags(1.0/(degrees_i+1e-7)).dot(mat_knn_ij) 
    
    mat_ij = mat_knn_ij.dot(mat_jj) # cell in mod i, gene in mod j
    
    if impute_j:
        return mat_ij, mat_ji
    else:
        return mat_ij

def impute_1pair(mod_i, mod_j, 
                 smoothed_features_i, smoothed_features_j,
                 settings,
                 knn, # 20
                 relaxation, # 3
                 output_knn_mat_ij='',
                 output_knn_mat_ji='',
                 impute_j=True,
                 ):
    """
    """
    # set up
    direct_i, direct_j = settings[mod_i].mod_direction, settings[mod_j].mod_direction
    
    mat_ii = smoothed_features_i.T # cell in mod i; gene in mod i
    mat_jj = smoothed_features_j.T # cell in mod j; gene in mod j
    
    genes_i = mat_ii.columns.values
    genes_j = mat_jj.columns.values
    genes_common = np.intersect1d(genes_i, genes_j)
    
    cells_i = mat_ii.index.values
    cells_j = mat_jj.index.values
    
    ## spearman correlation as distance  (rank -> zscore -> (flip sign?) -> "dot" distance) 
    # normalize the feature matrix
    mat_norm_i = (mat_ii[genes_common].rank(pct=True, axis=1)
                                      .apply(basic_utils.zscore, axis=1)
                                      *direct_i
                 )
    mat_norm_j = (mat_jj[genes_common].rank(pct=True, axis=1)
                                      .apply(basic_utils.zscore, axis=1)
                                      *direct_j
                 )
    maxk_i = int((len(cells_j)/len(cells_i))*knn*relaxation)+1 # max number of NN a cell in i can get 
    maxk_j = int((len(cells_i)/len(cells_j))*knn*relaxation)+1 # max number of NN a cell in j can get 
    
    if impute_j:
        # knn_i and knn_j
        # j <- i for each j, get kNN in i
        knn_ji = get_constrained_knn(mat_norm_j, mat_norm_i, knn=knn, k_saturate=maxk_i, metric='dot')
        mat_knn_ji = sparse_adj_to_mat(knn_ji, len(cells_j), len(cells_i))
        
        if output_knn_mat_ji:
            sparse.save_npz(output_knn_mat_ji, mat_knn_ji)
        
        # normalize 
        degrees_j = np.ravel(mat_knn_ji.sum(axis=1)) # for each cell in j, how many cells in i it connects to 
        mat_knn_ji = sparse.diags(1.0/(degrees_j+1e-7)).dot(mat_knn_ji) 
        
        # imputation both across and within modality
        mat_ji = mat_knn_ji.dot(mat_ii) # cell in mod j, gene in mod i
    
    
    # i <- j
    knn_ij = get_constrained_knn(mat_norm_i, mat_norm_j, knn=knn, k_saturate=maxk_j, metric='dot')
    mat_knn_ij = sparse_adj_to_mat(knn_ij, len(cells_i), len(cells_j))

    if output_knn_mat_ij:
        sparse.save_npz(output_knn_mat_ij, mat_knn_ij)
    
    degrees_i = np.ravel(mat_knn_ij.sum(axis=1)) # for each cell in i, how many cells in j it connects to 
    mat_knn_ij = sparse.diags(1.0/(degrees_i+1e-7)).dot(mat_knn_ij) 
    
    mat_ij = mat_knn_ij.dot(mat_jj) # cell in mod i, gene in mod j
    
    if impute_j:
        return mat_ij, mat_ji
    else:
        return mat_ij

def core_scf_routine(mods_selected, features_selected, settings, 
                    metas, gxc_hvftrs, 
                    ps, drop_npcs,
                    cross_mod_distance_measure, knn, relaxation, n_cca,
                    npc,
                    output_pcX_all, 
                    ka_smooth=5,
                    save_knn=False,
                    output_knn_within="",
                    output_knn_across="",
                    ):
    """smooth within modality, impute across modalities, and construct a joint PC matrix
    """
    # GENE * CELL !!!!
    smoothed_features = collections.OrderedDict()
    logging.info("Smoothing within modalities...")
    for mod in mods_selected:
        ti = time.time()
        if settings[mod].mod_category == 'mc':
            _df = gxc_hvftrs[mod]
        else:
            _mat = gxc_hvftrs[mod].data.todense()
            _df = pd.DataFrame(_mat, 
                              index=gxc_hvftrs[mod].gene, 
                              columns=gxc_hvftrs[mod].cell, 
                              ) 
        if len(metas[mod]) < 2:
            raise ValueError("Dataset too small (contains <2 cells)")

        npc = min(len(metas[mod]), npc)
        # added 08/03/2020
        ka = min(len(metas[mod]), ka_smooth) 
        k_smooth = min(len(metas[mod]), 3*ka)

        if save_knn: # added 07/27/2020
            output_knn_mat = output_knn_within.format(mod)
        else:
            output_knn_mat = ""

        # if k_smooth >= 2*ka:
        mat_smoothed, mat_knn = smooth_in_modality(_df, _df, k=k_smooth, ka=ka, npc=npc, 
                                                     p=ps[settings[mod].mod_category], 
                                                     drop_npc=drop_npcs[settings[mod].mod_category], 
                                                     output_filename=output_knn_mat,
                                                     )
        smoothed_features[mod] = mat_smoothed
        # else:
        #     smoothed_features[mod] = _df

        logging.info("{}: {}".format(mod, time.time()-ti))
    # delete
    del gxc_hvftrs[mod]

    # construct a joint matrix (PCA)
    logging.info("Constructing a joint matrix...")
    cells_all = np.hstack([metas[mod].index.values for mod in mods_selected]) # cell (all mods)  
    pcX_all = []
    for mod_y in features_selected: ## to 
        logging.info("Imputing into {} space...".format(mod_y))
        # get all_features
        X = []
        for mod_x in mods_selected:
            logging.info("for {} cells...".format(mod_x))
            if mod_x == mod_y:
                smoothed_yy = smoothed_features[mod_y].T # gene by cell !!! VERY IMPORTANT
                X.append(smoothed_yy)
            else:
                if save_knn: # added 07/27/2020
                    output_knn_mat_ij = output_knn_across.format(mod_x, mod_y)
                    output_knn_mat_ji = output_knn_across.format(mod_y, mod_x)
                else:
                    output_knn_mat_ij = "" 
                    output_knn_mat_ji = "" 
                # impute x cells y space
                smoothed_features_x = smoothed_features[mod_x]
                smoothed_features_y = smoothed_features[mod_y]
                if cross_mod_distance_measure == 'correlation':
                    imputed_xy = impute_1pair(mod_x, mod_y, 
                                              smoothed_features_x, smoothed_features_y,
                                              settings,
                                              knn=knn,
                                              relaxation=relaxation,
                                              impute_j=False,
                                              output_knn_mat_ij=output_knn_mat_ij,
                                              output_knn_mat_ji=output_knn_mat_ji,
                                              )
                elif cross_mod_distance_measure == 'cca':
                    imputed_xy = impute_1pair_cca(mod_x, mod_y, 
                                                 smoothed_features_x, smoothed_features_y,
                                                 settings,
                                                 knn=knn,
                                                 relaxation=relaxation,
                                                 n_cca=n_cca,
                                                 impute_j=False,
                                                 output_knn_mat_ij=output_knn_mat_ij,
                                                 output_knn_mat_ji=output_knn_mat_ji,
                                                )
                else:
                    raise ValueError("Choose from correlation and cca")
                X.append(imputed_xy)
        X = np.vstack(X) # cell (all mods) by gene (mod_y) 

        # # save X (imputed counts)
        # np.save(output_imputed_data_format.format(mod_y), X)

        # PCA
        U, s, V = fbpca.pca(X, npc)
        del X
        pcX = U.dot(np.diag(s))
        # normalize PCs
        sigma = np.sqrt(np.sum(s*s)/(pcX.shape[0]*pcX.shape[1]))
        pcX = pcX/sigma
        pcX_all.append(pcX)
        
    pcX_all = np.hstack(pcX_all)
    # save pcX_all
    adata_pcX_all = anndata.AnnData(
        X=pcX_all,
        obs=pd.DataFrame(index=cells_all),

    )
    adata_pcX_all.write(output_pcX_all)
    # np.save(output_pcX_all, pcX_all)
    # np.save(output_cells_all, cells_all)
    logging.info(f"Saved output to: {output_pcX_all}")
    return pcX_all, cells_all

def clustering_umap_routine(pcX_all, cells_all, mods_selected, metas, 
                            resolutions, k, 
                            umap_neighbors, min_dist, 
                            output_clst_and_umap,
                            cluster_only=False,
                            ):
    """
    """

    # clustering
    df_clsts = []
    for resolution in resolutions:
        logging.info('resolution r: {}'.format(resolution))
        df_clst = clst_utils.clustering_routine(
                                        pcX_all, 
                                        cells_all, k, 
                                        resolution=resolution,
                                        metric='euclidean', option='plain', n_trees=10, search_k=-1, verbose=False)
        df_clsts.append(df_clst.rename(columns={'cluster': 
                                                'cluster_joint_r{}'.format(resolution)
                                               }))
    df_clst = pd.concat(df_clsts, axis=1) 

    if cluster_only:
        # save results
        df_summary = df_clst
        df_summary['modality'] = ''
        for mod in mods_selected:
            _cells = metas[mod].index.values
            df_summary.loc[_cells, 'modality'] = mod
        df_summary.to_csv(output_clst_and_umap, sep='\t', header=True, index=True)    
        return df_summary

    else:
        # umap
        df_tsne = clst_utils.run_umap_lite(
                    pcX_all, 
                    cells_all, n_neighbors=umap_neighbors, min_dist=min_dist, n_dim=2, 
                    random_state=1)

        # summary
        df_summary = df_clst.join(df_tsne)
        df_summary['modality'] = ''
        for mod in mods_selected:
            _cells = metas[mod].index.values
            df_summary.loc[_cells, 'modality'] = mod
        # save results
        df_summary.to_csv(output_clst_and_umap, sep='\t', header=True, index=True)    
        return df_summary



"""
"""
import argparse
import logging

from robustlink.scf import SCF_main_repeat_subsampling
from robustlink import generate_metacells_rna
from robustlink import correlate_metacells_mc_rna
from robustlink import correlate_metacells_atac_rna

def create_parser():
    """
    """
    parser = argparse.ArgumentParser(prog='python -m robustlink')
    subparsers = parser.add_subparsers()

    # scfusion 
    scfusion = subparsers.add_parser('scfusion')
    scfusion.set_defaults(cmd='scfusion')
    SCF_main_repeat_subsampling.add_args(scfusion)

    # generat metacells rna
    metacell = subparsers.add_parser('metacell')
    metacell.set_defaults(cmd='metacell')
    generate_metacells_rna.add_args(metacell)

    # corr
    corr_mc = subparsers.add_parser('corr_mc')
    corr_mc.set_defaults(cmd='corr_mc')
    correlate_metacells_mc_rna.add_args(corr_mc)

    # corr
    corr_atac = subparsers.add_parser('corr_atac')
    corr_atac.set_defaults(cmd='corr_atac')
    correlate_metacells_atac_rna.add_args(corr_atac)

    return parser


if __name__ == "__main__":
    # 
    parser = create_parser()
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)

    if not hasattr(args, 'cmd'):
        raise ValueError("Wrong input, please run: `python -m robustlink --help`")
    elif args.cmd == 'scfusion':
        SCF_main_repeat_subsampling.main(args)
    elif args.cmd == 'metacell':
        generate_metacells_rna.main(args)
    elif args.cmd == 'corr_mc':
        correlate_metacells_mc_rna.main(args)
    elif args.cmd == 'corr_atac':
        correlate_metacells_atac_rna.main(args)


#!/usr/bin/env python
# coding: utf-8

import os
from scipy import sparse
import argparse
import logging
import pandas as pd
import numpy as np
import anndata

from robustlink import utils
from robustlink import enhancer_gene_utils

logger = utils.create_logger()

def pipe_corr_analysis_mc(
        common_rna_cells, common_mc_cells,
        cell_cell_knn_xaxis, cell_cell_knn_yaxis,
        common_genes,
        X, Y_cg, Y_mcg, 
        modx_clsts, knn_xy, 
        enhancer_gene_to_eval,
        output_corrs,
        corr_type='pearsonr',
        force=False,
        num_metacell_limit=0,
        num_metacell_limit_low=0,
        shuff_enhs=False,
    ):
    """
    """
    # new cells  
    common_rna_cells_updated = np.intersect1d(common_rna_cells, cell_cell_knn_xaxis)
    common_mc_cells_updated = np.intersect1d(common_mc_cells, cell_cell_knn_yaxis)

    # make sure the original matrices have the correct index
    x_idx = utils.get_index_from_array(common_rna_cells, common_rna_cells_updated)
    y_idx = utils.get_index_from_array(common_mc_cells, common_mc_cells_updated)
    X = X.tocsc()[:, x_idx] 
    Y_cg = Y_cg.tocsc()[:, y_idx]
    Y_mcg = Y_mcg.tocsc()[:, y_idx] 

    # make sure knn_xy, knn_xx have the right cell index
    cell_idx_xaxis = utils.get_index_from_array(cell_cell_knn_xaxis, common_rna_cells_updated)
    cell_idx_yaxis = utils.get_index_from_array(cell_cell_knn_yaxis, common_mc_cells_updated)
    knn_xy = knn_xy.tocsr()[cell_idx_xaxis,:].tocsc()[:,cell_idx_yaxis] # x-by-y
    modx_clsts = modx_clsts.reindex(common_rna_cells_updated)

    logging.info("{}_{}_{}_{}_{}".format(knn_xy.shape, modx_clsts.shape, X.shape, Y_cg.shape, Y_mcg.shape))

    for clst_col in modx_clsts.columns: 
        logging.info(clst_col)
        output_corr = output_corrs.format(clst_col)
        if not force and os.path.isfile(output_corr):
            logging.info("skip {}, already exists...".format(output_corr))
            continue # skip the existing file

        # choose one clustering to proceed
        uniq_labels = np.sort(modx_clsts[clst_col].unique()) 
        logging.info("Number of metacells: {}".format(len(uniq_labels)))
        if num_metacell_limit > 0 and len(uniq_labels) > num_metacell_limit:
            logging.info("skip {}, exceeding max num_metacell_limit...".format(len(uniq_labels)))
            continue
        if num_metacell_limit_low > 0 and len(uniq_labels) < num_metacell_limit_low:
            logging.info("skip {}, below min num_metacell_limit...".format(len(uniq_labels)))
            continue

        knn_xz = enhancer_gene_utils.turn_cluster_labels_to_knn(modx_clsts[clst_col].values, 
                                            uniq_labels,
                                           )

        # gene by metacell (counts)
        gc_rna = X.dot(knn_xz).todense() 
        # normalization (logCPM)
        gc_rna = utils.logcpm(pd.DataFrame(gc_rna)).values

        # enhancer by metacell (counts cg, mcg)
        knn_yz = knn_xy.T.dot(knn_xz)
        ec_cg = Y_cg.dot(knn_yz).todense() 
        ec_mcg = Y_mcg.dot(knn_yz).todense()  
        logging.info("{} {} {}".format(gc_rna.shape, ec_cg.shape, ec_mcg.shape))

        # mC
        ec_mccg = utils.get_mcc_lite_v4(
                                       pd.DataFrame(ec_cg).astype(np.float32), 
                                       pd.DataFrame(ec_mcg).astype(np.float32), 
                                       base_call_cutoff=5, sufficient_coverage_fraction=0.8, fillna=True)
        logging.info("{}".format(ec_mccg.shape))

        # corr analysis
        output = enhancer_gene_utils.compute_enh_gene_corrs(
            gc_rna, ec_mccg, 
            common_genes, ec_mccg.index.values,
            enhancer_gene_to_eval['gene'].values, 
            enhancer_gene_to_eval['enh'].values, 
            output_file="", corr_type=corr_type, chunksize=100000, verbose_level=0,
            shuff_enhs=shuff_enhs,
            )

        # save results
        if len(output) == 4:
            (to_correlate, corrs, corrs_shuffled, corrs_shuffled_cells) = output
            res_corrs = enhancer_gene_to_eval[to_correlate].copy()
            res_corrs['corr'] = corrs 
            res_corrs['corr_shuff'] = corrs_shuffled
            res_corrs['corr_shuff_cells'] = corrs_shuffled_cells
            res_corrs.to_csv(output_corr, sep='\t', header=True, index=True)

        elif len(output) == 6:
            (to_correlate, corrs, corrs_shuffled, corrs_shuffled_cells, 
                corrs_shuffled_enhs,
                corrs_shuffled_enhs_bygroups,
                ) = output
            res_corrs = enhancer_gene_to_eval[to_correlate].copy()
            res_corrs['corr'] = corrs 
            res_corrs['corr_shuff'] = corrs_shuffled
            res_corrs['corr_shuff_cells'] = corrs_shuffled_cells
            res_corrs['corr_shuff_enhs'] = corrs_shuffled_enhs
            res_corrs['corr_shuff_enhs_bygroups'] = corrs_shuffled_enhs_bygroups
            res_corrs.to_csv(output_corr, sep='\t', header=True, index=True)

    return output # last of the kind

def wrap_corr_analysis_mc(
        out_dir,
        f_egpairs, f_gene, f_enh,
        scfusion_dir, mod_x, mod_y, input_name_tag, i_sub,
        corr_type='pearsonr',
        force=False,
        num_metacell_limit=0,
        num_metacell_limit_low=0,
        shuff_enhs=False,
        save_todisk=True,
    ):
    """
    """
    # output: (i, k, --r)
    if save_todisk:
        output_corrs = os.path.join(out_dir, f'{input_name_tag}_s{i_sub}_corr_{corr_type}_{{}}.tsv') 
                              # cluster resolution to be filled in
    else:
        output_corrs = ''

    ### input file paths
    # # input enh-gene tables, gene-by-cell, enhancer-by-cell matrices
    # f_egpairs = os.path.join(data_dir, 'enhancer_gene_pairs_1mbp.tsv')
    # f_gene    = os.path.join(data_dir, 'counts_gene_rna.h5ad')
    # f_enh     = os.path.join(data_dir, 'counts_enh_mc.h5ad')

    # for knn_xx
    input_modx_clsts =       [os.path.join(scfusion_dir, f'{input_name_tag}_s{i_sub}_metacells_{mod_x}.tsv.gz')]
    # for knn_xy
    input_knn_cells_xaxis  =  os.path.join(scfusion_dir, f'{input_name_tag}_s{i_sub}_cells_{mod_x}.txt')
    input_knn_cells_yaxis  =  os.path.join(scfusion_dir, f'{input_name_tag}_s{i_sub}_cells_{mod_y}.txt')
    input_knn_xy =            os.path.join(scfusion_dir, f'{input_name_tag}_s{i_sub}_knn_across_{mod_x}_{mod_y}.npz') 
    ### end of file paths

    # # Load data 
    # load enhancer-gene linkage
    enhancer_gene_to_eval = pd.read_csv(f_egpairs, sep='\t')

    # load count matrices
    adata_gene = anndata.read(f_gene)
    adata_enh  = anndata.read(f_enh)

    common_rna_cells = adata_gene.var.index.values
    common_genes = adata_gene.obs.index.values
    X = adata_gene.X

    common_mc_cells  = adata_enh.var.index.values 
    common_enhancer_regions = adata_enh.obs # check if this works
    Y_mcg = adata_enh.layers['mcg']
    Y_cg = adata_enh.layers['cg']

    # load knn networks 
    # for knn_xx 
    modx_clsts = pd.concat([
        pd.read_csv(fname, sep='\t',index_col=0)
        for fname in input_modx_clsts
    ], axis=1)
    # for knn_xy 
    knn_xy = sparse.load_npz(input_knn_xy)  
    cell_cell_knn_xaxis = np.loadtxt(input_knn_cells_xaxis, dtype=str)
    cell_cell_knn_yaxis = np.loadtxt(input_knn_cells_yaxis, dtype=str)

    logging.info("{} {} {} {}".format(
            modx_clsts.shape, 
            knn_xy.shape, 
            cell_cell_knn_xaxis.shape, 
            cell_cell_knn_yaxis.shape,
            )
            )

    pipe_corr_analysis_mc(
        common_rna_cells, common_mc_cells,
        cell_cell_knn_xaxis, cell_cell_knn_yaxis,
        common_genes,
        X, Y_cg, Y_mcg, 
        modx_clsts, knn_xy, 
        enhancer_gene_to_eval,
        output_corrs,
        corr_type=corr_type,
        force=force,
        num_metacell_limit=num_metacell_limit,
        num_metacell_limit_low=num_metacell_limit_low,
        shuff_enhs=shuff_enhs,
    )
    return 

def add_args(parser):
    """
    """
    parser.add_argument('--tolink',         required=True, type=str, help='a table of enhancer-gene pairs to be assessed for linkages')
    parser.add_argument('--countdata_gene', required=True, type=str, help='gene by cell count matrix; h5ad format')
    parser.add_argument('--countdata_enh',  required=True, type=str, help='enhancer by cell count matrix; h5ad format')
    parser.add_argument('-o', '--out_dir',  required=True, type=str)

    # to pull scFusion results (kNN graphs)
    parser.add_argument('--scfusion_dir',  required=True)
    parser.add_argument('--fusiondata_rna',  required=True, type=str, help='data for scFusion (h5ad file) -- need the filename to pull scFusion results')
    parser.add_argument('--fusiondata_mc',   required=True, type=str, help='data for scFusion (h5ad file) -- need the filename to pull scFusion results')
    parser.add_argument('-tag',  '--input_name_tag', required=True, type=str)
    parser.add_argument('-isub', '--i_sub', type=str, help="[0-9]")

    parser.add_argument('-ct', '--corr_type', choices=['pearsonr', 'spearmanr'],  
                                              default='pearsonr',
                                              help="choose from pearsonr or spearmanr")
    parser.add_argument('-f', '--force', action='store_true', help='to overwrite existing file')
    parser.add_argument('-n', '--num_metacell_limit', default=0, type=int, help='max num of metacells')
    return 

def main(args):
    # output setting
    # run this with each combination of (i_sub, knn)
    f_egpairs = args.tolink
    f_gene = args.countdata_gene
    f_enh  = args.countdata_enh

    out_dir = args.out_dir
    scfusion_dir = args.scfusion_dir

    mod_x = os.path.basename(args.fusiondata_rna).replace('.h5ad', '')
    mod_y = os.path.basename(args.fusiondata_mc ).replace('.h5ad', '')

    input_name_tag = args.input_name_tag
    i_sub = args.i_sub
    corr_type = args.corr_type
    force = args.force
    num_metacell_limit = args.num_metacell_limit

    logging.info(",".join([mod_x, mod_y, input_name_tag, str(i_sub), corr_type, str(force), str(num_metacell_limit)]))

    # run
    wrap_corr_analysis_mc(
        out_dir,
        f_egpairs, f_gene, f_enh,
        scfusion_dir, mod_x, mod_y, input_name_tag, i_sub,
        corr_type=corr_type,
        force=force,
        num_metacell_limit=num_metacell_limit,
    )

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    add_args(parser)
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    main(args)